#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from titus.datatype import *
from titus.signature import *
import titus.P as P

class TestSignature(unittest.TestCase):
    version = PFAVersion(0, 8, 1)

    def typeEquality(self, x, y):
        if isinstance(x, AvroType) and isinstance(y, AvroType):
            return x.accepts(y) and y.accepts(x)
        elif isinstance(x, FcnType) and isinstance(y, FcnType):
            return all(xt.accepts(yt) and yt.accepts(xt) for xt, yt in zip(x.params, y.params)) and x.ret.accepts(y.ret) and y.ret.accepts(x.ret)
        else:
            return False

    def matches(self, x, y):
        if x is None and y is None:
            return True
        elif x is None or y is None:
            return False
        else:
            (dummy, xparams, xret) = x
            (yparams, yret) = y
            return all(self.typeEquality(xp, yp) for xp, yp in zip(xparams, yparams)) and self.typeEquality(xret, yret)
            
    def testPassThroughExactParameterMatches(self):
        self.assertTrue(self.matches(Sig([{"x": P.Null()}], P.Null()).accepts([AvroNull()], self.version), ([AvroNull()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroInt()], self.version), ([AvroInt()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroLong()], self.version), ([AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroFloat()], self.version), ([AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroDouble()], self.version), ([AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Bytes()}], P.Null()).accepts([AvroBytes()], self.version), ([AvroBytes()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.String()}], P.Null()).accepts([AvroString()], self.version), ([AvroString()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroInt())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroInt())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Union([P.Int()])}], P.Null()).accepts([AvroUnion([AvroInt()])], self.version), ([AvroUnion([AvroInt()])], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Fixed(10)}], P.Null()).accepts([AvroFixed(10, "MyFixed")], self.version), ([AvroFixed(10, "MyFixed")], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Fixed(10, "MyFixed")}], P.Null()).accepts([AvroFixed(10, "MyFixed")], self.version), ([AvroFixed(10, "MyFixed")], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Enum(["one", "two", "three"])}], P.Null()).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), ([AvroEnum(["one", "two", "three"], "MyEnum")], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Enum(["one", "two", "three"], "MyEnum")}], P.Null()).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), ([AvroEnum(["one", "two", "three"], "MyEnum")], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Record({"one": P.Int()})}], P.Null()).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Record({"one": P.Int()}, "MyRecord")}], P.Null()).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], AvroNull())))

    def testNotMatchAntiPatterns(self):
        self.assertTrue(self.matches(Sig([{"x": P.Null()}], P.Null()).accepts([AvroInt()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroLong()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroFloat()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroDouble()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroBytes()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Bytes()}], P.Null()).accepts([AvroString()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.String()}], P.Null()).accepts([AvroArray(AvroInt())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroUnion([AvroInt()])], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Union([P.Int()])}], P.Null()).accepts([AvroFixed(10, "MyFixed")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Fixed(10, "YourFixed")}], P.Null()).accepts([AvroFixed(10, "MyFixed")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Fixed(10)}], P.Null()).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Enum(["one", "two", "three"], "YourEnum")}], P.Null()).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Enum(["one", "two", "three"])}], P.Null()).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Record({"one": P.Int()}, "YourRecord")}], P.Null()).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Record({"one": P.Int()})}], P.Null()).accepts([AvroNull()], self.version), None))

    def testPromoteNumbers(self):
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroInt()], self.version), ([AvroInt()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroLong()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroFloat()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Int()}], P.Null()).accepts([AvroDouble()], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroInt()], self.version), ([AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroLong()], self.version), ([AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroFloat()], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Long()}], P.Null()).accepts([AvroDouble()], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroInt()], self.version), ([AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroLong()], self.version), ([AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroFloat()], self.version), ([AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Float()}], P.Null()).accepts([AvroDouble()], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroInt()], self.version), ([AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroLong()], self.version), ([AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroDouble()], self.version), ([AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Double()}], P.Null()).accepts([AvroDouble()], self.version), ([AvroDouble()], AvroNull())))

        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroInt())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroArray(AvroLong())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroArray(AvroFloat())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Int())}], P.Null()).accepts([AvroArray(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Long())}], P.Null()).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroLong())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Long())}], P.Null()).accepts([AvroArray(AvroLong())], self.version), ([AvroArray(AvroLong())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Long())}], P.Null()).accepts([AvroArray(AvroFloat())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Long())}], P.Null()).accepts([AvroArray(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Float())}], P.Null()).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Float())}], P.Null()).accepts([AvroArray(AvroLong())], self.version), ([AvroArray(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Float())}], P.Null()).accepts([AvroArray(AvroFloat())], self.version), ([AvroArray(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Float())}], P.Null()).accepts([AvroArray(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Double())}], P.Null()).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Double())}], P.Null()).accepts([AvroArray(AvroLong())], self.version), ([AvroArray(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Double())}], P.Null()).accepts([AvroArray(AvroDouble())], self.version), ([AvroArray(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Double())}], P.Null()).accepts([AvroArray(AvroDouble())], self.version), ([AvroArray(AvroDouble())], AvroNull())))

        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroInt())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroMap(AvroLong())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroMap(AvroFloat())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Int())}], P.Null()).accepts([AvroMap(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Long())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroLong())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Long())}], P.Null()).accepts([AvroMap(AvroLong())], self.version), ([AvroMap(AvroLong())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Long())}], P.Null()).accepts([AvroMap(AvroFloat())], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Long())}], P.Null()).accepts([AvroMap(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Float())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Float())}], P.Null()).accepts([AvroMap(AvroLong())], self.version), ([AvroMap(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Float())}], P.Null()).accepts([AvroMap(AvroFloat())], self.version), ([AvroMap(AvroFloat())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Float())}], P.Null()).accepts([AvroMap(AvroDouble())], self.version), None))

        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Double())}], P.Null()).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Double())}], P.Null()).accepts([AvroMap(AvroLong())], self.version), ([AvroMap(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Double())}], P.Null()).accepts([AvroMap(AvroDouble())], self.version), ([AvroMap(AvroDouble())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Map(P.Double())}], P.Null()).accepts([AvroMap(AvroDouble())], self.version), ([AvroMap(AvroDouble())], AvroNull())))

    def testLooselyMatchFunctionReferences(self):
        self.assertTrue(self.matches(Sig([{"x": P.Fcn([P.Long()], P.Long())}], P.Null()).accepts([FcnType([AvroLong()], AvroLong())], self.version), ([FcnType([AvroLong()], AvroLong())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Fcn([P.Long()], P.Long())}], P.Null()).accepts([FcnType([AvroLong()], AvroInt())], self.version), ([FcnType([AvroLong()], AvroInt())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Fcn([P.Long()], P.Long())}], P.Null()).accepts([FcnType([AvroDouble()], AvroLong())], self.version), ([FcnType([AvroDouble()], AvroLong())], AvroNull())))

    def testMatchWildcards(self):
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroNull()], self.version), ([AvroNull()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroInt()], self.version), ([AvroInt()], AvroInt())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroLong()], self.version), ([AvroLong()], AvroLong())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroFloat()], self.version), ([AvroFloat()], AvroFloat())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroDouble()], self.version), ([AvroDouble()], AvroDouble())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroBytes()], self.version), ([AvroBytes()], AvroBytes())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroString()], self.version), ([AvroString()], AvroString())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroInt())], AvroArray(AvroInt()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroInt())], AvroMap(AvroInt()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroUnion([AvroInt()])], self.version), ([AvroUnion([AvroInt()])], AvroUnion([AvroInt()]))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroFixed(10, "MyFixed")], self.version), ([AvroFixed(10, "MyFixed")], AvroFixed(10, "MyFixed"))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), ([AvroEnum(["one", "two", "three"], "MyEnum")], AvroEnum(["one", "two", "three"], "MyEnum"))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Wildcard("A")).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], AvroRecord([AvroField("one", AvroInt())], "MyRecord"))))

    def testMatchNestedWildcards(self):
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroNull())], self.version), ([AvroArray(AvroNull())], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroInt())], AvroInt())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroLong())], self.version), ([AvroArray(AvroLong())], AvroLong())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroFloat())], self.version), ([AvroArray(AvroFloat())], AvroFloat())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroDouble())], self.version), ([AvroArray(AvroDouble())], AvroDouble())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroBytes())], self.version), ([AvroArray(AvroBytes())], AvroBytes())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroString())], self.version), ([AvroArray(AvroString())], AvroString())))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroArray(AvroInt()))], self.version), ([AvroArray(AvroArray(AvroInt()))], AvroArray(AvroInt()))))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroMap(AvroInt()))], self.version), ([AvroArray(AvroMap(AvroInt()))], AvroMap(AvroInt()))))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroUnion([AvroInt()]))], self.version), ([AvroArray(AvroUnion([AvroInt()]))], AvroUnion([AvroInt()]))))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroFixed(10, "MyFixed"))], self.version), ([AvroArray(AvroFixed(10, "MyFixed"))], AvroFixed(10, "MyFixed"))))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroEnum(["one", "two", "three"], "MyEnum"))], self.version), ([AvroArray(AvroEnum(["one", "two", "three"], "MyEnum"))], AvroEnum(["one", "two", "three"], "MyEnum"))))
        self.assertTrue(self.matches(Sig([{"x": P.Array(P.Wildcard("A"))}], P.Wildcard("A")).accepts([AvroArray(AvroRecord([AvroField("one", AvroInt())], "MyRecord"))], self.version), ([AvroArray(AvroRecord([AvroField("one", AvroInt())], "MyRecord"))], AvroRecord([AvroField("one", AvroInt())], "MyRecord"))))

        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroNull()], self.version), ([AvroNull()], AvroArray(AvroNull()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroInt()], self.version), ([AvroInt()], AvroArray(AvroInt()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroLong()], self.version), ([AvroLong()], AvroArray(AvroLong()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroFloat()], self.version), ([AvroFloat()], AvroArray(AvroFloat()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroDouble()], self.version), ([AvroDouble()], AvroArray(AvroDouble()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroBytes()], self.version), ([AvroBytes()], AvroArray(AvroBytes()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroString()], self.version), ([AvroString()], AvroArray(AvroString()))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroArray(AvroInt())], self.version), ([AvroArray(AvroInt())], AvroArray(AvroArray(AvroInt())))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroMap(AvroInt())], self.version), ([AvroMap(AvroInt())], AvroArray(AvroMap(AvroInt())))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroUnion([AvroInt()])], self.version), ([AvroUnion([AvroInt()])], AvroArray(AvroUnion([AvroInt()])))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroFixed(10, "MyFixed")], self.version), ([AvroFixed(10, "MyFixed")], AvroArray(AvroFixed(10, "MyFixed")))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroEnum(["one", "two", "three"], "MyEnum")], self.version), ([AvroEnum(["one", "two", "three"], "MyEnum")], AvroArray(AvroEnum(["one", "two", "three"], "MyEnum")))))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A")}], P.Array(P.Wildcard("A"))).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], AvroArray(AvroRecord([AvroField("one", AvroInt())], "MyRecord")))))

    def testUseWildcardsToNormalizeNumericalTypes(self):
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroInt(), AvroInt()], self.version), ([AvroInt(), AvroInt()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroInt(), AvroLong()], self.version), ([AvroLong(), AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroInt(), AvroFloat()], self.version), ([AvroFloat(), AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroInt(), AvroDouble()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))

        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroLong(), AvroInt()], self.version), ([AvroLong(), AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroLong(), AvroLong()], self.version), ([AvroLong(), AvroLong()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroLong(), AvroFloat()], self.version), ([AvroFloat(), AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroLong(), AvroDouble()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))

        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroFloat(), AvroInt()], self.version), ([AvroFloat(), AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroFloat(), AvroLong()], self.version), ([AvroFloat(), AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroFloat(), AvroFloat()], self.version), ([AvroFloat(), AvroFloat()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroFloat(), AvroDouble()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))

        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroDouble(), AvroInt()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroDouble(), AvroLong()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroDouble(), AvroFloat()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))
        self.assertTrue(self.matches(Sig([{"x": P.Wildcard("A", set([AvroInt(), AvroLong(), AvroFloat(), AvroDouble()]))}, {"y": P.Wildcard("A")}], P.Null()).accepts([AvroDouble(), AvroDouble()], self.version), ([AvroDouble(), AvroDouble()], AvroNull())))

    def testMatchWildRecords(self):
        self.assertTrue(self.matches(Sig([{"x": P.WildRecord("A", {"one": P.Int()})}], P.Wildcard("A")).accepts([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt())], "MyRecord")], AvroRecord([AvroField("one", AvroInt())], "MyRecord"))))
        self.assertTrue(self.matches(Sig([{"x": P.WildRecord("A", {"one": P.Int()})}], P.Wildcard("A")).accepts([AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroString())], "MyRecord")], self.version), ([AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroString())], "MyRecord")], AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroString())], "MyRecord"))))
        self.assertTrue(self.matches(Sig([{"x": P.WildRecord("A", {"one": P.String()})}], P.Wildcard("A")).accepts([AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroString())], "MyRecord")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.WildRecord("A", {"one": P.Double()})}], P.Wildcard("A")).accepts([AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroString())], "MyRecord")], self.version), None))
        self.assertTrue(self.matches(Sig([{"x": P.WildRecord("A", {"one": P.Int()})}], P.Wildcard("A")).accepts([AvroRecord([AvroField("uno", AvroInt()), AvroField("two", AvroString())], "MyRecord")], self.version), None))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast as pythonast
import re

import ply.lex as lex
import ply.yacc as yacc

import titus.producer.tools as t

class ParserError(Exception):
    """Exception for errors encountered in parsing a pfainspector command line."""
    def __init__(self, text, col):
        self.text = text
        self.col = col
        super(ParserError, self).__init__("syntax error:  {0}\n--------------{1}^".format(text, "-" * col))
    def __repr__(self):
        return "ParserError({0}, {1})".format(self.text, self.col)

class Ast(object):
    """Abstract syntax tree for a pfainspector command line.

    Allows for the possibility that a command line is incomplete (for tab-completion).
    """
    partial = False
    col = None
    def checkPartial(self, text):
        if self.partial:
            raise ParserError(text, self.col)
    def value(self):
        raise TypeError("not a pure value: {0}".format(self))
    def regex(self):
        return self.value()
    def replacement(self):
        return self.value()

class DotDotDot(Ast):
    """AST for "..."."""
    def __init__(self, col=None):
        self.col = col
    def __repr__(self):
        return "DotDotDot()"
    def __str__(self):
        return "..."
    def __eq__(self, other):
        return isinstance(other, DotDotDot)

class Underscore(Ast):
    """AST for "_" (meaning a regular expression wildcard)."""
    def __init__(self, col=None):
        self.col = col
    def __repr__(self):
        return "Underscore()"
    def __str__(self):
        return "_"
    def __eq__(self, other):
        return isinstance(other, Underscore)
    def regex(self):
        return t.Any()

class PlusOrMinus(Ast):
    """AST for "+-"."""
    def __init__(self, col=None):
        self.col = col
    def __repr__(self):
        return "PlusOrMinus()"
    def __str__(self):
        return "+-"
    def __eq__(self, other):
        return isinstance(other, PlusOrMinus)

class FilePath(Ast):
    """AST for a filesystem file path."""
    def __init__(self, text, col=None):
        self.text = text
        self.col = col
    def __repr__(self):
        return "FilePath({0})".format(repr(self.text))
    def __str__(self):
        return self.text
    def __eq__(self, other):
        return isinstance(other, FilePath) and self.text == other.text

class Regex(Ast):
    """AST for a JSON regular expression."""
    def __init__(self, text, find, to, flags, col=None):
        self.text = text
        self.find = find
        self.to = to
        self.flags = flags
        self.col = col
    def __repr__(self):
        return "Regex({0}, {1}, {2}, {3})".format(repr(self.text), repr(self.find), repr(self.to), repr(self.flags))
    def __str__(self):
        return self.text
    def __eq__(self, other):
        return isinstance(other, Regex) and self.text == other.text
    def regex(self):
        return t.RegEx(self.find, self.to, self.flags)

class Word(FilePath):
    """AST for a space-delimited word in a command line."""
    def __init__(self, text, col=None):
        self.text = text
        self.col = col
    def __repr__(self):
        return "Word({0})".format(repr(self.text))
    def __str__(self):
        return self.text
    def __eq__(self, other):
        return isinstance(other, Word) and self.text == other.text
    def value(self):
        if self.text == "null":
            return None
        elif self.text == "true":
            return True
        elif self.text == "false":
            return False
        else:
            return self.text

class StartExtract(Ast):
    """AST for the beginning of an extraction (pulling subobjects out of JSON)."""
    def __init__(self, text, col=None):
        self.text = text
        self.col = col
    def __repr__(self):
        return "StartExtract({0})".format(repr(self.text))
    def __str__(self):
        return self.text
    def __eq__(self, other):
        return isinstance(other, StartExtract) and self.text == other.text

class String(FilePath):
    """AST for a string."""
    def __init__(self, text, partial, col=None):
        self.text = text
        self.partial = partial
        self.col = col
    def __repr__(self):
        return "String({0}{1})".format(repr(self.text), " partial" if self.partial else "")
    def __str__(self):
        return repr(self.text)
    def __eq__(self, other):
        return isinstance(other, String) and self.text == other.text
    def value(self):
        return self.text

class Switches(FilePath):
    """AST for argument switches (arguments beginning with "-", as on a UNIX command line)."""
    def __init__(self, text, col=None):
        self.text = text
        self.col = col
    def __repr__(self):
        return "Switches({0})".format(repr(self.text))
    def __str__(self):
        return self.text
    def __eq__(self, other):
        return isinstance(other, Switches) and self.text == other.text

class Floating(Ast):
    """AST for a floating-point number."""
    def __init__(self, num, col=None):
        self.num = float(num)
        self.col = col
    def __repr__(self):
        return "Floating({0})".format(repr(self.num))
    def __str__(self):
        return str(self.num)
    def __eq__(self, other):
        return isinstance(other, Floating) and self.num == other.num
    def value(self):
        return self.num

class Integer(Floating):
    """AST for an integer."""
    def __init__(self, num, col=None):
        self.num = int(num)
        self.col = col
    def __repr__(self):
        return "Integer({0})".format(repr(self.num))
    def __str__(self):
        return str(self.num)
    def __eq__(self, other):
        return isinstance(other, Integer) and self.num == other.num
    def value(self):
        return self.num

class JsonObject(Ast):
    """AST for a JSON object, which may be partial (no closing curly bracket) and strings need not be quoted."""
    def __init__(self, pairs, partial, ellipsis, col=None):
        self.pairs = pairs
        self.partial = partial
        self.ellipsis = ellipsis
        self.col = col
    def __repr__(self):
        return "JsonObject([{0}]{1}{2})".format(", ".join(map(repr, self.pairs)), " partial" if self.partial else "", " ellipsis" if self.ellipsis else "")
    def __str__(self):
        return "{" + ", ".join(map(str, self.pairs)) + (" ..." if self.ellipsis else "") + "}"
    def __eq__(self, other):
        return isinstance(other, JsonObject) and self.pairs == other.pairs
    def checkPartial(self, text):
        super(JsonObject, self).checkPartial(text)
        for pair in self.pairs:
            pair.checkPartial(text)
    def value(self):
        if self.ellipsis:
            super(JsonObject, self).value()
        else:
            return dict((p.key.value(), p.value.value()) for p in self.pairs)
    def regex(self):
        if self.ellipsis:
            return t.Min(**dict((p.key.value(), p.value.regex()) for p in self.pairs))
        else:
            return dict((p.key.value(), p.value.regex()) for p in self.pairs)
    def replacement(self):
        if self.ellipsis:
            super(JsonObject, self).replacement()
        else:
            return dict((p.key.replacement(), p.value.replacement()) for p in self.pairs)

class Pair(Ast):
    """AST for a key-value pair in a JSON object."""
    def __init__(self, key, value, partial, col=None):
        self.key = key
        self.value = value
        self.partial = partial
        self.col = col
    def __repr__(self):
        return "Pair({0}, {1}{2})".format(repr(self.key), repr(self.value), " partial" if self.partial else "")
    def __str__(self):
        return self.key + ": " + str(self.value)
    def __eq__(self, other):
        return isinstance(other, Pair) and self.key == other.key and self.value == other.value
    def checkPartial(self, text):
        super(Pair, self).checkPartial(text)
        self.key.checkPartial(text)
        self.value.checkPartial(text)

class Approx(Ast):
    """AST for "~" (meaning approximate)."""
    def __init__(self, central, error, col=None):
        self.central = central
        self.error = error
        self.col = col
    def __repr__(self):
        return "Approx({0}, {1})".format(repr(self.central), repr(self.error))
    def __str__(self):
        return str(self.central) + " +- " + str(self.error)
    def __eq__(self):
        return isinstance(other, Approx) and self.central == other.central and self.error == other.error
    def regex(self):
        return t.Approx(self.central, self.error)

class Replacement(object):
    """AST for a JSON regular expression replacement."""
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return "Replacement(" + self.name + ")"
    def __str__(self):
        return "(" + self.name + ")"

class Group(Ast):
    """AST for a JSON regular expression group."""
    def __init__(self, item, counter, col=None):
        self.item = item
        self.counter = counter
        self.col = col
    def __repr__(self):
        return "Group({0}, {1})".format(repr(self.item), self.counter)
    def __str__(self):
        return "(" + str(self.item) + ")"
    def __eq__(self):
        return isinstance(other, Group) and self.item == other.item
    def regex(self):
        return t.Group(**{str(self.counter): self.item.regex()})
    def replacement(self):
        if isinstance(self.item, AndChain) and len(self.item.chain) == 1 and isinstance(self.item.chain[0], Integer):
            return Replacement(str(self.item.chain[0].num))
        else:
            super(Group, self).replacement()

class Range(Ast):
    """AST for a JSON regular expression numerical range."""
    def __init__(self, operator, comparator, col=None):
        self.operator = operator
        self.comparator = comparator
        self.col = col
    def __repr__(self):
        return "Range({0} {1})".format(self.operator, self.comparator)
    def __str__(self):
        return "# " + self.operator + " " + str(self.comparator)
    def __eq__(self):
        return isinstance(other, Range) and self.operator == other.operator and self.comparator == other.comparator
    def regex(self):
        if self.operator is None:
            return t.Any(int, int, float)
        elif self.operator == "<":
            return t.LT(self.comparator)
        elif self.operator == "<=":
            return t.LE(self.comparator)
        elif self.operator == ">":
            return t.GT(self.comparator)
        elif self.operator == ">=":
            return t.GE(self.comparator)
        else:
            raise ValueError

class OrChain(Ast):
    """AST for a chain of logical-or expressions (vertical bar) in a JSON regular expression."""
    def __init__(self, chain, col=None):
        self.chain = chain
        self.col = col
    def __repr__(self):
        return "OrChain([{0}])".format(", ".join(map(repr, self.chain)))
    def __str__(self):
        return " | ".join(map(str, self.chain))
    def __eq__(self):
        return isinstance(other, OrChain) and self.chain == other.chain
    def regex(self):
        return t.Or(*[x.regex() for x in self.chain])

class AndChain(Ast):
    """AST for a chain of logical-and expressions (ampersand) in a JSON regular expression."""
    def __init__(self, chain, col=None):
        self.chain = chain
        self.col = col
    def __repr__(self):
        return "AndChain([{0}])".format(", ".join(map(repr, self.chain)))
    def __str__(self):
        return " & ".join(map(str, self.chain))
    def __eq__(self):
        return isinstance(other, AndChain) and self.chain == other.chain
    def regex(self):
        return t.And(*[x.regex() for x in self.chain])

class JsonArray(Ast):
    """AST for a JSON array, which may be partial (no closing square bracket) and strings need not be quoted."""
    def __init__(self, items, partial, startEllipsis, endEllipsis, col=None):
        self.items = items
        self.partial = partial
        self.startEllipsis = startEllipsis
        self.endEllipsis = endEllipsis
        self.col = col
    def __repr__(self):
        return "JsonArray([{0}]{1}{2}{3})".format(", ".join(map(repr, self.items)), " partial" if self.partial else "", " startEllipsis" if self.startEllipsis else "", " endEllipsis" if self.endEllipsis else "")
    def __str__(self):
        return "[" + ("... " if self.startEllipsis else "") + ", ".join(map(str, self.items)) + (" ..." if self.endEllipsis else "") + "]"
    def __eq__(self, other):
        return isinstance(other, JsonArray) and self.items == other.items
    def checkPartial(self, text):
        super(JsonArray, self).checkPartial(text)
        for item in self.items:
            item.checkPartial(text)
    def value(self):
        if self.startEllipsis or self.endEllipsis:
            super(JsonObject, self).value()
        else:
            return [x.value() for x in self.items]
    def value(self):
        if self.startEllipsis and self.endEllipsis:
            raise NotImplementedError
        elif self.startEllipsis and not self.endEllipsis:
            return t.End(*[x.regex() for x in self.items])
        elif not self.startEllipsis and self.endEllipsis:
            return t.Start(*[x.regex() for x in self.items])
        else:
            return [x.regex() for x in self.items]
    def replacement(self):
        if self.startEllipsis or self.endEllipsis:
            super(JsonObject, self).replacement()
        else:
            return [x.replacement() for x in self.items]

class Extract(Ast):
    """AST for an extraction from a JSON/PFA file in memory."""
    def __init__(self, text, items, partial, col=None):
        self.text = text
        self.items = items
        self.partial = partial
        self.col = col
    def __repr__(self):
        return "Extract({0}, [{1}]{2})".format(repr(self.text), ", ".join(map(repr, self.items)), " partial" if self.partial else "")
    def __str__(self):
        return str(self.text) + "[" + ", ".join(map(str, self.items)) + "]"
    def strto(self, index):
        return str(self.text) + "[" + ", ".join(map(str, self.items[:index])) + "]"
    def __eq__(self, other):
        return isinstance(other, Extract) and self.text == other.text and self.items == other.items
    def checkPartial(self, text):
        super(Extract, self).checkPartial(text)
        for item in self.items:
            item.checkPartial(text)

class Option(Ast):
    """AST for a command argument (name=value)."""
    def __init__(self, word, value, partial, col=None):
        self.word = word
        self.value = value
        self.partial = partial
        self.col = col
    def __repr__(self):
        return "Option({0}, {1}{2})".format(repr(self.word), repr(self.value), " partial" if self.partial else "")
    def __str__(self):
        return str(self.word) + "=" + str(self.value)
    def __eq__(self, other):
        return isinstance(other, Option) and self.word == other.word and self.value == other.value
    def checkPartial(self, text):
        super(Option, self).checkPartial(text)
        self.word.checkPartial(text)
        self.value.checkPartial(text)

class Parser(object):
    """Parser for the "ply" package, specialized for pfainspector command lines.

    Includes both the tokenizer and the parser.
    """

    def __init__(self):
        tokens = ["LT", "LE", "GT", "GE", "DOTDOTDOT", "UNDERSCORE", "PLUSORMINUS", "REGEX", "REGEXSUBS", "FILEPATH", "WORD", "STARTEXTRACT", "STRING", "PARTSTRING", "INTEGER", "FLOATING", "SWITCHES"]
        literals = [",", ":", "[", "]", "{", "}", "=", "(", ")", "|", "&", "#"]

        def t_LE(t):
            r'''<='''
            return t

        def t_LT(t):
            r'''<'''
            return t

        def t_GE(t):
            r'''>='''
            return t

        def t_GT(t):
            r'''>'''
            return t

        def t_DOTDOTDOT(t):
            r'''\.\.\.'''
            t.value = DotDotDot(t.lexpos)
            return t

        def t_UNDERSCORE(t):
            r'''_'''
            t.value = Underscore(t.lexpos)
            return t

        def t_PLUSORMINUS(t):
            r'''\+\-'''
            t.value = PlusOrMinus(t.lexpos)
            return t

        def t_STARTEXTRACT(t):
            r'''[A-Za-z][A-Za-z0-9_\.]*\['''
            t.value = StartExtract(t.value.rstrip("["), t.lexpos)
            return t

        def t_REGEX(t):
            r'''\?(/(([^\\/]|\\.)*)/)([ilmsuxILMSUX]+)?'''
            m = re.match(t_REGEX.__doc__, t.value)
            find = m.groups()[1]
            flags = m.groups()[3]
            t.value = Regex(t.value, find, None, flags, t.lexpos)
            return t

        def t_REGEXSUBS(t):
            r'''@(/(([^\\/]|\\.)*)/)((([^\\/]|\\.)*)/)([ilmsuxILMSUX]+)?'''
            m = re.match(t_REGEXSUBS.__doc__, t.value)
            find = m.groups()[1]
            to = m.groups()[4]
            flags = m.groups()[6]
            t.value = Regex(t.value, find, to, flags, t.lexpos)
            return t

        def t_FILEPATH(t):
            r'''[A-Za-z0-9_\.~@$%\-\+\\]*(/[A-Za-z0-9_\.~@$%\-\+\\]*)+'''
            t.value = FilePath(t.value, t.lexpos)
            return t

        def t_WORD(t):
            r'''[A-Za-z][A-Za-z0-9_\.]*'''
            t.value = Word(t.value, t.lexpos)
            return t

        def t_STRING(t):
            r'''"([^\\"]|\\.)*"'''
            t.value = String(pythonast.literal_eval(t.value), False, t.lexpos)
            return t

        def t_PARTSTRING(t):
            r'''"([^\\"]|\\.)*$'''
            t.value = String(pythonast.literal_eval(t.value), True, t.lexpos)
            return t

        def t_FLOATING(t):
            r'''-?(([0-9]+\.[0-9]*|\.[0-9]+)([eE][-+]?[0-9]+)?|([0-9]+(\.[0-9]*)?|\.[0-9]+)[eE][-+]?[0-9]+)'''
            t.value = Floating(t.value, t.lexpos)
            return t

        def t_INTEGER(t):
            r'''-?[0-9]+'''
            t.value = Integer(t.value, t.lexpos)
            return t

        def t_SWITCHES(t):
            r'''-[A-Za-z0-9]+'''
            t.value = Switches(t.value, t.lexpos)
            return t

        t_ignore = " \t"

        def t_error(t):
            raise ParserError(self.text, t.lexpos)

        self.lexer = lex.lex()

        def p_expressions(p):
            r'''expressions : expression
                            | expressions expression'''
            if len(p) == 2:
                p[0] = [p[1]]
            else:
                if p[1][-1].partial:
                    raise ParserError(self.text, self.lexer.lexpos)
                p[1].append(p[2])
                p[0] = p[1]

        def p_expression(p):
            r'''expression : value
                           | extract
                           | option
                           | FILEPATH
                           | SWITCHES'''
            p[0] = p[1]

        def p_option(p):
            r'''option : WORD "="
                       | WORD "=" value'''
            p[0] = Option(p[1], p[3] if len(p) == 4 else None, len(p) < 4, self.lexer.lexpos)

        def p_extract(p):
            r'''extract : STARTEXTRACT
                        | STARTEXTRACT indexes
                        | STARTEXTRACT indexes "]"'''
            p[0] = Extract(p[1].text, p[2] if len(p) > 2 else [], len(p) < 4, self.lexer.lexpos)

        def p_index(p):
            r'''index : WORD
                      | STRING
                      | PARTSTRING
                      | INTEGER'''
            p[0] = p[1]

        def p_indexes(p):
            r'''indexes : index
                        | indexes ","
                        | indexes "," index'''
            if len(p) == 2:
                p[0] = [p[1]]
            elif len(p) == 3:
                p[0] = p[1]
            else:
                p[1].append(p[3])
                p[0] = p[1]

        def p_value(p):
            r'''value : WORD
                      | STRING
                      | PARTSTRING
                      | INTEGER
                      | FLOATING
                      | UNDERSCORE
                      | REGEX
                      | REGEXSUBS
                      | jsonobject
                      | jsonarray
                      | range
                      | approx
                      | group'''
            p[0] = p[1]

        def p_values(p):
            r'''values : empty
                       | value
                       | values "," value'''
            if len(p) == 2:
                if p[1] is None:
                    p[0] = []
                else:
                    p[0] = [p[1]]
            else:
                if len(p[1]) > 0:
                    prev = p[1][-1]
                    if prev.partial:
                        raise ParserError(self.text, self.lexer.lexpos)
                p[1].append(p[3])
                p[0] = p[1]

        def p_approx(p):
            r'''approx : INTEGER PLUSORMINUS INTEGER
                       | INTEGER PLUSORMINUS FLOATING
                       | FLOATING PLUSORMINUS INTEGER
                       | FLOATING PLUSORMINUS FLOATING'''
            p[0] = Approx(p[1].num, p[3].num, self.lexer.lexpos)
            
        def p_group(p):
            r'''group : "(" orchain ")"
                      | "(" andchain ")"'''
            p[0] = Group(p[2], self.groupCounter, self.lexer.lexpos)
            self.groupCounter += 1

        def p_range1(p):
            r'''range : "#"'''
            p[0] = Range(None, None, self.lexer.lexpos)

        def p_range2(p):
            r'''range : "#" LT INTEGER
                      | "#" LT FLOATING
                      | "#" LE INTEGER
                      | "#" LE FLOATING
                      | "#" GT INTEGER
                      | "#" GT FLOATING
                      | "#" GE INTEGER
                      | "#" GE FLOATING'''
            p[0] = Range(p[2], p[3].num, self.lexer.lexpos)

        def p_range3(p):
            r'''range : INTEGER  LT "#" LT INTEGER
                      | INTEGER  LT "#" LT FLOATING
                      | FLOATING LT "#" LT INTEGER
                      | FLOATING LT "#" LT FLOATING
                      | INTEGER  LT "#" LE INTEGER
                      | INTEGER  LT "#" LE FLOATING
                      | FLOATING LT "#" LE INTEGER
                      | FLOATING LT "#" LE FLOATING
                      | INTEGER  LE "#" LT INTEGER
                      | INTEGER  LE "#" LT FLOATING
                      | FLOATING LE "#" LT INTEGER
                      | FLOATING LE "#" LT FLOATING
                      | INTEGER  LE "#" LE INTEGER
                      | INTEGER  LE "#" LE FLOATING
                      | FLOATING LE "#" LE INTEGER
                      | FLOATING LE "#" LE FLOATING'''
            reverse = {"<": ">=", "<=": ">"}
            p[0] = AndChain([Range(reverse[p[2]], p[1].num, self.lexer.lexpos),
                             Range(p[4], p[5].num, self.lexer.lexpos)])

        def p_orchain(p):
            r'''orchain : andchain
                        | orchain "|" andchain'''
            if len(p) == 2:
                p[0] = OrChain([p[1]], self.lexer.lexpos)
            else:
                p[1].chain.append(p[3])
                p[1].col = self.lexer.lexpos
                p[0] = p[1]

        def p_andchain(p):
            r'''andchain : value
                         | andchain "&" value'''
            if len(p) == 2:
                p[0] = AndChain([p[1]], self.lexer.lexpos)
            else:
                p[1].chain.append(p[3])
                p[1].col = self.lexer.lexpos
                p[0] = p[1]

        def p_jsonarray(p):
            r'''jsonarray : jsonarraypart "]"
                          | jsonarraypart DOTDOTDOT "]"'''
            p[1].partial = False
            if isinstance(p[2], DotDotDot):
                if p[1].startEllipsis:
                    raise ParserError(self.text, self.lexer.lexpos)
                p[1].endEllipsis = True
            p[0] = p[1]

        def p_jsonarraypart(p):
            r'''jsonarraypart : "[" values
                              | "[" DOTDOTDOT values
                              | "[" values ","
                              | "[" DOTDOTDOT values ","'''
            if isinstance(p[2], DotDotDot):
                p[0] = JsonArray(p[3], partial=True, startEllipsis=True, endEllipsis=False, col=self.lexer.lexpos)
            else:
                p[0] = JsonArray(p[2], partial=True, startEllipsis=False, endEllipsis=False, col=self.lexer.lexpos)

        def p_jsonobject(p):
            r'''jsonobject : jsonobjectpart "}"
                           | jsonobjectpart DOTDOTDOT "}"'''
            p[1].partial = False
            if isinstance(p[2], DotDotDot):
                if p[1].ellipsis:
                    raise ParserError(self.text, self.lexer.lexpos)
                p[1].ellipsis = True
            p[0] = p[1]

        def p_jsonobjectpart(p):
            r'''jsonobjectpart : "{" pairs
                               | "{" DOTDOTDOT pairs
                               | "{" pairs ","
                               | "{" DOTDOTDOT pairs ","'''
            if isinstance(p[2], DotDotDot):
                p[0] = JsonObject(p[3], partial=True, ellipsis=True, col=self.lexer.lexpos)
            else:
                p[0] = JsonObject(p[2], partial=True, ellipsis=False, col=self.lexer.lexpos)

        def p_pairs(p):
            r'''pairs : empty
                      | pair
                      | pairs "," pair'''
            if len(p) == 2:
                if p[1] is None:
                    p[0] = []
                else:
                    p[0] = [p[1]]
            else:
                if len(p[1]) > 0:
                    if p[1][-1].partial:
                        raise ParserError(self.text, self.lexer.lexpos)
                p[1].append(p[3])
                p[0] = p[1]

        def p_pair(p):
            r'''pair : WORD ":"
                     | STRING ":"
                     | WORD ":" value
                     | STRING ":" value'''
            if len(p) < 4:
                p[0] = Pair(p[1], None, True, self.lexer.lexpos)
            else:
                p[0] = Pair(p[1], p[3], False, self.lexer.lexpos)

        def p_empty(p):
            r'''empty : '''

        def p_error(p):
            raise ParserError(self.text, self.lexer.lexpos)

        self.yacc = yacc.yacc(debug=False, write_tables=False)

    def parse(self, text):
        """Parse the given text, returning an abstract syntax tree.

        :type text: string
        :param text: command line to parse
        :rtype: titus.inspector.parser.Ast
        :return: parsed text as an abstract syntax tree
        """

        self.text = text
        self.groupCounter = 1
        return self.yacc.parse(text, lexer=self.lexer)

parser = Parser()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.signature import Lifespan
from titus.signature import PFAVersion
from titus.datatype import *
from titus.errors import PFARuntimeException
from titus.util import callfcn, negativeIndex, startEnd
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "s."

#################################################################### basic access

class Len(LibFcn):
    name = prefix + "len"
    sig = Sig([{"s": P.String()}], P.Int())
    errcodeBase = 39000
    def genpy(self, paramTypes, args, pos):
        return "len({0})".format(*args)
    def __call__(self, state, scope, pos, paramTypes, s):
        return len(s)
provide(Len())

class Substr(LibFcn):
    name = prefix + "substr"
    sig = Sig([{"s": P.String()}, {"start": P.Int()}, {"end": P.Int()}], P.String())
    errcodeBase = 39010
    def __call__(self, state, scope, pos, paramTypes, s, start, end):
        return s[start:end]
provide(Substr())

class SubstrTo(LibFcn):
    name = prefix + "substrto"
    sig = Sig([{"s": P.String()}, {"start": P.Int()}, {"end": P.Int()}, {"replacement": P.String()}], P.String())
    errcodeBase = 39020
    def __call__(self, state, scope, pos, paramTypes, s, start, end, replacement):
        normStart, normEnd = startEnd(len(s), start, end)
        before = s[:normStart]
        after = s[normEnd:]
        return before + replacement + after
provide(SubstrTo())

#################################################################### searching

class Contains(LibFcn):
    name = prefix + "contains"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Boolean())
    errcodeBase = 39030
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        return needle in haystack
provide(Contains())

class Count(LibFcn):
    name = prefix + "count"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Int())
    errcodeBase = 39040
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        if len(haystack) == 0 or len(needle) == 0:
            return 0
        else:
            return haystack.count(needle)
provide(Count())

class Index(LibFcn):
    name = prefix + "index"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Int())
    errcodeBase = 39050
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        try:
            return haystack.index(needle)
        except ValueError:
            return -1
provide(Index())

class RIndex(LibFcn):
    name = prefix + "rindex"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Int())
    errcodeBase = 39060
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        try:
            return haystack.rindex(needle)
        except ValueError:
            return -1
provide(RIndex())

class StartsWith(LibFcn):
    name = prefix + "startswith"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Boolean())
    errcodeBase = 39070
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        return haystack.startswith(needle)
provide(StartsWith())

class EndsWith(LibFcn):
    name = prefix + "endswith"
    sig = Sig([{"haystack": P.String()}, {"needle": P.String()}], P.Boolean())
    errcodeBase = 39080
    def __call__(self, state, scope, pos, paramTypes, haystack, needle):
        return haystack.endswith(needle)
provide(EndsWith())

#################################################################### conversions to/from other types

class Join(LibFcn):
    name = prefix + "join"
    sig = Sig([{"array": P.Array(P.String())}, {"sep": P.String()}], P.String())
    errcodeBase = 39090
    def __call__(self, state, scope, pos, paramTypes, array, sep):
        return sep.join(array)
provide(Join())

class Split(LibFcn):
    name = prefix + "split"
    sig = Sig([{"s": P.String()}, {"sep": P.String()}], P.Array(P.String()))
    errcodeBase = 39100
    def __call__(self, state, scope, pos, paramTypes, s, sep):
        if len(sep) == 0:
            return []
        else:
            return s.split(sep)
provide(Split())

class Hex(LibFcn):
    name = prefix + "hex"
    sig = Sigs([Sig([{"x": P.Long()}], P.String()),
                Sig([{"x": P.Long()}, {"width": P.Int()}, {"zeroPad": P.Boolean()}], P.String())])
    errcodeBase = 39110
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(args) == 1:
            x, = args
            if x < 0:
                raise PFARuntimeException("negative number", self.errcodeBase + 1, self.name, pos)
            else:
                return "{0:x}".format(x)

        else:
            x, width, zeroPad = args
            if x < 0:
                raise PFARuntimeException("negative number", self.errcodeBase + 1, self.name, pos)
            if not zeroPad:
                if width < 0:
                    formatStr = "{0:<" + str(-width) + "x}"
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:" + str(width) + "x}"
            else:
                if width < 0:
                    raise PFARuntimeException("negative width cannot be used with zero-padding", self.errcodeBase + 0, self.name, pos)
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:0" + str(width) + "x}"
            return formatStr.format(x)
provide(Hex())
        
class StringInt(LibFcn):
    name = prefix + "int"
    sig = Sigs([Sig([{"x": P.Long()}], P.String()),
                Sig([{"x": P.Long()}, {"width": P.Int()}, {"zeroPad": P.Boolean()}], P.String())])
    errcodeBase = 39240
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(args) == 1:
            x, = args
            return "{0:d}".format(x)

        elif len(args) == 3:
            x, width, zeroPad = args
            if not zeroPad:
                if width < 0:
                    formatStr = "{0:<" + str(-width) + "d}"
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:" + str(width) + "d}"
            else:
                if width < 0:
                    raise PFARuntimeException("negative width cannot be used with zero-padding", self.errcodeBase + 0, self.name, pos)
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:0" + str(width) + "d}"
            return formatStr.format(x)
provide(StringInt())

class Number(LibFcn):
    name = prefix + "number"
    sig = Sigs([Sig([{"x": P.Long()}], P.String(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use s.int for integers")),
                Sig([{"x": P.Long()}, {"width": P.Int()}, {"zeroPad": P.Boolean()}], P.String(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use s.int for integers")),
                Sig([{"x": P.Double()}, {"width": P.Union([P.Int(), P.Null()])}, {"precision": P.Union([P.Int(), P.Null()])}], P.String()),
                Sig([{"x": P.Double()}, {"width": P.Union([P.Int(), P.Null()])}, {"precision": P.Union([P.Int(), P.Null()])}, {"minNoExp": P.Double()}, {"maxNoExp": P.Double()}], P.String())])
    errcodeBase = 39120
    def __call__(self, state, scope, pos, paramTypes, *args):
        if len(args) == 1 and paramTypes[0] in ("int", "long"):
            x, = args
            return "{0:d}".format(x)

        elif len(args) == 3 and paramTypes[0] in ("int", "long"):
            x, width, zeroPad = args
            if not zeroPad:
                if width < 0:
                    formatStr = "{0:<" + str(-width) + "d}"
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:" + str(width) + "d}"
            else:
                if width < 0:
                    raise PFARuntimeException("negative width cannot be used with zero-padding", self.errcodeBase + 0, self.name, pos)
                elif width == 0:
                    formatStr = ""
                else:
                    formatStr = "{0:0" + str(width) + "d}"
            return formatStr.format(x)

        else:
            if len(args) == 3:
                x, width, precision = args
                minNoExp, maxNoExp = 0.0001, 100000
            else:
                x, width, precision, minNoExp, maxNoExp = args

            if width is None:
                widthStr = ""
            else:
                if isinstance(width, dict):
                    width = width["int"]
                if width < 0:
                    widthStr = "<" + str(-width)
                else:
                    widthStr = str(width)

            if precision is None:
                precisionStr = ".6"
            else:
                if isinstance(precision, dict):
                    precision = precision["int"]
                if precision < 0:
                    raise PFARuntimeException("negative precision", self.errcodeBase + 1, self.name, pos)
                else:
                    precisionStr = "." + str(precision)

            if x == 0.0: x = 0.0   # drop sign bit from zero

            v = abs(x)
            if v == 0.0 or (v >= minNoExp and v <= maxNoExp):
                conv = "f"
            else:
                conv = "e"

            formatStr = "{0:" + widthStr + precisionStr + conv + "}"
            result = formatStr.format(x)

            if precision is None:
                m = re.search("\.[0-9]+?(0+) *$", result)
                if m is None:
                    m = re.search("\.[0-9]+?(0+)e", result)
                if m is not None:
                    start, end = m.regs[1]
                    numChanged = end - start
                    result = result[:start] + result[end:]
                    if width is not None:
                        if width < 0:
                            actual, target = len(result), -width
                            if actual < target:
                                result = result + (" " * (target - actual))
                        elif width == 0:
                            result = ""
                        else:
                            actual, target = len(result), width
                            if actual < target:
                                result = (" " * (target - actual)) + result

            return result

provide(Number())

#################################################################### conversions to/from other strings

class Concat(LibFcn):
    name = prefix + "concat"
    sig = Sig([{"x": P.String()}, {"y": P.String()}], P.String())
    errcodeBase = 39130
    def __call__(self, state, scope, pos, paramTypes, x, y):
        return x + y
provide(Concat())

class Repeat(LibFcn):
    name = prefix + "repeat"
    sig = Sig([{"s": P.String()}, {"n": P.Int()}], P.String())
    errcodeBase = 39140
    def __call__(self, state, scope, pos, paramTypes, s, n):
        return s * n
provide(Repeat())

class Lower(LibFcn):
    name = prefix + "lower"
    sig = Sig([{"s": P.String()}], P.String())
    errcodeBase = 39150
    def __call__(self, state, scope, pos, paramTypes, s):
        return s.lower()
provide(Lower())

class Upper(LibFcn):
    name = prefix + "upper"
    sig = Sig([{"s": P.String()}], P.String())
    errcodeBase = 39160
    def __call__(self, state, scope, pos, paramTypes, s):
        return s.upper()
provide(Upper())

class LStrip(LibFcn):
    name = prefix + "lstrip"
    sig = Sig([{"s": P.String()}, {"chars": P.String()}], P.String())
    errcodeBase = 39170
    def __call__(self, state, scope, pos, paramTypes, s, chars):
        return s.lstrip(chars)
provide(LStrip())

class RStrip(LibFcn):
    name = prefix + "rstrip"
    sig = Sig([{"s": P.String()}, {"chars": P.String()}], P.String())
    errcodeBase = 39180
    def __call__(self, state, scope, pos, paramTypes, s, chars):
        return s.rstrip(chars)
provide(RStrip())

class Strip(LibFcn):
    name = prefix + "strip"
    sig = Sig([{"s": P.String()}, {"chars": P.String()}], P.String())
    errcodeBase = 39190
    def __call__(self, state, scope, pos, paramTypes, s, chars):
        return s.strip(chars)
provide(Strip())

class ReplaceAll(LibFcn):
    name = prefix + "replaceall"
    sig = Sig([{"s": P.String()}, {"original": P.String()}, {"replacement": P.String()}], P.String())
    errcodeBase = 39200
    def __call__(self, state, scope, pos, paramTypes, s, original, replacement):
        return s.replace(original, replacement)
provide(ReplaceAll())

class ReplaceFirst(LibFcn):
    name = prefix + "replacefirst"
    sig = Sig([{"s": P.String()}, {"original": P.String()}, {"replacement": P.String()}], P.String())
    errcodeBase = 39210
    def __call__(self, state, scope, pos, paramTypes, s, original, replacement):
        if len(s) > 0:
            return s.replace(original, replacement, 1)
        else:
            return s.replace(original, replacement)
provide(ReplaceFirst())

class ReplaceLast(LibFcn):
    name = prefix + "replacelast"
    sig = Sig([{"s": P.String()}, {"original": P.String()}, {"replacement": P.String()}], P.String())
    errcodeBase = 39220
    def __call__(self, state, scope, pos, paramTypes, s, original, replacement):
        backward = "".join(reversed(s))
        if len(s) > 0:
            replaced = backward.replace("".join(reversed(original)), "".join(reversed(replacement)), 1)
        else:
            replaced = backward.replace("".join(reversed(original)), "".join(reversed(replacement)))
        return "".join(reversed(replaced))
provide(ReplaceLast())

class Translate(LibFcn):
    name = prefix + "translate"
    sig = Sig([{"s": P.String()}, {"oldchars": P.String()}, {"newchars": P.String()}], P.String())
    errcodeBase = 39230
    def __call__(self, state, scope, pos, paramTypes, s, oldchars, newchars):
        out = []
        for c in s:
            try:
                i = oldchars.index(c)
            except ValueError:
                out.append(c)
            else:
                if i < len(newchars):
                    out.append(newchars[i])
        return "".join(out)
provide(Translate())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn
from titus.util import div
import titus.P as P

def np():
    import numpy
    return numpy

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "interp."

class Bin(LibFcn):
    name = prefix + "bin"
    sig = Sigs([Sig([{"x": P.Double()}, {"numbins": P.Int()}, {"low": P.Double()}, {"high": P.Double()}], P.Int()),
                Sig([{"x": P.Double()}, {"origin": P.Double()}, {"width": P.Double()}], P.Int())])
    errcodeBase = 22000
    def __call__(self, state, scope, pos, paramTypes, x, *args):
        if len(args) == 3:
            numbins, low, high = args
            if low >= high or math.isnan(low) or math.isnan(high):
                raise PFARuntimeException("bad histogram range", self.errcodeBase + 0, self.name, pos)
            if numbins < 1:
                raise PFARuntimeException("bad histogram scale", self.errcodeBase + 1, self.name, pos)
            if math.isnan(x) or x < low or x >= high:
                raise PFARuntimeException("x out of range", self.errcodeBase + 2, self.name, pos)

            out = int(math.floor(numbins * div((x - low), (high - low))))

            if out < 0 or out >= numbins:
                raise PFARuntimeException("x out of range", self.errcodeBase + 2, self.name, pos)
            return out
        else:
            origin, width = args
            if math.isnan(origin) or math.isinf(origin):
                raise PFARuntimeException("bad histogram range", self.errcodeBase + 0, self.name, pos)
            if width <= 0.0 or math.isnan(width):
                raise PFARuntimeException("bad histogram scale", self.errcodeBase + 1, self.name, pos)
            if math.isnan(x) or math.isinf(x):
                raise PFARuntimeException("x out of range", self.errcodeBase + 2, self.name, pos)
            else:
                return int(math.floor(div((x - origin), width)))
provide(Bin())
    
class Nearest(LibFcn):
    name = prefix + "nearest"
    sig = Sigs([Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Wildcard("T")}))}], P.Wildcard("T")),
                Sig([{"x": P.Array(P.Double())}, {"table": P.Array(P.WildRecord("R", {"x": P.Array(P.Double()), "to": P.Wildcard("T")}))}], P.Wildcard("T")),
                Sig([{"x": P.Wildcard("X1")}, {"table": P.Array(P.WildRecord("R", {"x": P.Wildcard("X2"), "to": P.Wildcard("T")}))}, {"metric": P.Fcn([P.Wildcard("X1"), P.Wildcard("X2")], P.Double())}], P.Wildcard("T"))])
    errcodeBase = 22010
    def __call__(self, state, scope, pos, paramTypes, datum, table, *metric):
        if len(table) == 0:
            raise PFARuntimeException("table must have at least one entry", self.errcodeBase + 0, self.name, pos)
        if len(metric) == 1:
            metric, = metric
            # do signature 3
            one = None
            oned = None
            for item in table:
                d = callfcn(state, scope, metric, [datum, item["x"]])
                if one is None or d < oned:
                    one = item
                    oned = d
            return one["to"] 
        elif isinstance(paramTypes[0], dict) and paramTypes[0].get("type") == "array":
            # do signature 2
            one = None
            oned = None
            for item in table:
                x = item["x"]
                if len(x) != len(datum):
                    raise PFARuntimeException("inconsistent dimensionality", self.errcodeBase + 1, self.name, pos)
                d = sum((x0i - xi)**2 for x0i, xi in zip(datum, x))
                if one is None or d < oned:
                    one = item
                    oned = d
            return one["to"]
        else:
            # do signature 1
            one = None
            oned = None
            for item in table:
                d = abs(datum - item["x"])
                if one is None or d < oned:
                    one = item
                    oned = d
            return one["to"]
provide(Nearest())

class Linear(LibFcn):
    name = prefix + "linear"
    sig = Sigs([Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Double()}))}], P.Double()),
                Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Array(P.Double())}))}], P.Array(P.Double()))])
    errcodeBase = 22020
    @staticmethod
    def closest(datum, table, code, fcnName, pos):
        below = None
        above = None
        belowd = None
        aboved = None
        for item in table:
            if item["x"] <= datum and (below is None or datum - item["x"] < belowd):
                below = item
                belowd = datum - item["x"]
            if datum < item["x"] and (above is None or item["x"] - datum < aboved):
                above = item
                aboved = item["x"] - datum
        if below is not None and above is not None:
            return below, above, True
        else:
            one = None
            two = None
            oned = None
            twod = None
            for item in table:
                d = abs(datum - item["x"])
                if one is None or d < oned:
                    two = one
                    twod = oned
                    one = item
                    oned = d
                elif (two is None or d < twod) and d != oned:
                    two = item
                    twod = d
            if two is None:
                raise PFARuntimeException("table must have at least two distinct x values", code, fcnName, pos)
            return one, two, False
    @staticmethod
    def interpolateSingle(datum, one, two):
        onex = one["x"]
        twox = two["x"]
        oney = one["to"]
        twoy = two["to"]
        unitless = div((datum - onex), (twox - onex))
        return (1.0 - unitless)*oney + unitless*twoy
    @staticmethod
    def interpolateMulti(datum, one, two, code, fcnName, pos):
        onex = one["x"]
        twox = two["x"]
        oney = one["to"]
        twoy = two["to"]
        unitless = div((datum - onex), (twox - onex))
        if len(oney) != len(twoy):
            raise PFARuntimeException("inconsistent dimensionality", code, fcnName, pos)
        return [(1.0 - unitless)*oney[i] + unitless*twoy[i] for i in range(len(oney))]
    def __call__(self, state, scope, pos, paramTypes, datum, table):
        one, two, between = Linear.closest(datum, table, self.errcodeBase + 0, self.name, pos)
        if isinstance(paramTypes[-1], dict) and paramTypes[-1].get("type") == "array":
            return Linear.interpolateMulti(datum, one, two, self.errcodeBase + 1, self.name, pos)
        else:
            return Linear.interpolateSingle(datum, one, two)
provide(Linear())

class LinearFlat(LibFcn):
    name = prefix + "linearFlat"
    sig = Sigs([Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Double()}))}], P.Double()),
                Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Array(P.Double())}))}], P.Array(P.Double()))])
    errcodeBase = 22030
    def __call__(self, state, scope, pos, paramTypes, datum, table):
        one, two, between = Linear.closest(datum, table, self.errcodeBase + 0, self.name, pos)
        if not between:
            return one["to"]
        elif isinstance(paramTypes[-1], dict) and paramTypes[-1].get("type") == "array":
            return Linear.interpolateMulti(datum, one, two, self.errcodeBase + 1, self.name, pos)
        else:
            return Linear.interpolateSingle(datum, one, two)
provide(LinearFlat())

class LinearMissing(LibFcn):
    name = prefix + "linearMissing"
    sig = Sigs([Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Double()}))}], P.Union([P.Null(), P.Double()])),
                Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Array(P.Double())}))}], P.Union([P.Null(), P.Array(P.Double())]))])
    errcodeBase = 22040
    def __call__(self, state, scope, pos, paramTypes, datum, table):
        one, two, between = Linear.closest(datum, table, self.errcodeBase + 0, self.name, pos)
        if not between and one["x"] != datum:
            return None
        elif isinstance([x for x in paramTypes[-1] if x != "null"][0], dict) and [x for x in paramTypes[-1] if x != "null"][0].get("type") == "array":
            return Linear.interpolateMulti(datum, one, two, self.errcodeBase + 1, self.name, pos)
        else:
            return Linear.interpolateSingle(datum, one, two)
provide(LinearMissing())

# barycentric (could be a 2-d version with triangles or an arbitrary-dimension one with hypervolumes)
# bilinear (requires a 2-d grid)
# trilinear (requires a 3-d grid)
# spline (also 1-d)
# cubic (also 1-d)
# loess (also 1-d)



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

def np():
    import numpy
    return numpy

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import div
from titus.lib.core import INT_MIN_VALUE, INT_MAX_VALUE, LONG_MIN_VALUE, LONG_MAX_VALUE
import titus.P as P
from titus.lib.prob.dist import Chi2Distribution

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "stat.test."

class KSTwoSample(LibFcn):
    name = prefix + "kolmogorov"
    sig = Sig([{"x": P.Array(P.Double())}, {"y": P.Array(P.Double())}], P.Double())
    errcodeBase = 38000
    def __call__(self, state, scope, pos, paramTypes, x, y):
        x = sorted([xi for xi in x if not math.isnan(xi)])
        y = sorted([yi for yi in y if not math.isnan(yi)])
        n1, n2 = len(x), len(y)
        if (x == y):
            return 1.0
        elif ((len(x) == 0) or (len(y) == 0)):
            return 0.0
        else:
            j1 = j2 = 0
            fn1 = fn2 = d = 0.0
            while ((j1 < n1) and (j2 < n2)):
                d1 = x[j1]
                d2 = y[j2]
                if d1 <= d2:
                    j1 += 1
                    fn1 = float(j1)/n1
                if d2 <= d1:
                    j2 += 1
                    fn2 = float(j2)/n2
                dt = abs(fn2 - fn1)
                if dt > d:
                    d = dt
            en = math.sqrt((n1 * n2)/float(n1 + n2))
            stat = (en + 0.12 + 0.11/en)*d
            return 1.0 - kolomogorov_cdf(stat)
provide(KSTwoSample())

def kolomogorov_cdf(z):
    """
    Function to compute the cdf of the Kolomogorov distribution.  Used by other
        variants of KS test
    """
    L = math.sqrt(2.0 * math.pi) * (1.0/z)
    sumpart = 0.0
    for v in range(1, 150):
        sumpart += math.exp(-((2.0*v - 1.0)**2)*((math.pi**2)/(8.0*(z**2))))
    return L*sumpart

class Residual(LibFcn):
    name = prefix + "residual"
    sig = Sigs([Sig([{"observation": P.Double()}, {"prediction": P.Double()}], P.Double()),
                Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())}], P.Array(P.Double())),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())}], P.Map(P.Double()))])
    errcodeBase = 38010
    def __call__(self, state, scope, pos, paramTypes, observation, prediction):
        if isinstance(observation, dict):
            if set(observation.keys()) != set(prediction.keys()):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            return dict((k, observation[k] - prediction[k]) for k in observation)

        elif isinstance(observation, (tuple, list)):
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            return [float(o - p) for o, p in zip(observation, prediction)]

        else:
            return float(observation - prediction)

provide(Residual())

class Pull(LibFcn):
    name = prefix + "pull"
    sig = Sigs([Sig([{"observation": P.Double()}, {"prediction": P.Double()}, {"uncertainty": P.Double()}], P.Double()),
                Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())}, {"uncertainty": P.Array(P.Double())}], P.Array(P.Double())),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())}, {"uncertainty": P.Map(P.Double())}], P.Map(P.Double()))])
    errcodeBase = 38020
    def __call__(self, state, scope, pos, paramTypes, observation, prediction, uncertainty):
        if isinstance(observation, dict):
            if set(observation.keys()) != set(prediction.keys()):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            if set(observation.keys()) != set(uncertainty.keys()):
                raise PFARuntimeException("misaligned uncertainty", self.errcodeBase + 1, self.name, pos)
            return dict((k, div(observation[k] - prediction[k], uncertainty[k])) for k in observation)

        elif isinstance(observation, (tuple, list)):
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            if len(observation) != len(uncertainty):
                raise PFARuntimeException("misaligned uncertainty", self.errcodeBase + 1, self.name, pos)
            return [float(div(o - p, u)) for o, p, u in zip(observation, prediction, uncertainty)]

        else:
            return div(observation - prediction, uncertainty)

provide(Pull())

class Mahalanobis(LibFcn):
    name = prefix + "mahalanobis"
    sig = Sigs([Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())},
                     {"covariance": P.Array(P.Array(P.Double()))}], P.Double()),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())},
                     {"covariance": P.Map(P.Map(P.Double()))}], P.Double())])
    errcodeBase = 38030
    def __call__(self, state, scope, pos, paramTypes, observation, prediction, covariance):
        if isinstance(observation, (tuple, list)):
            if len(observation) < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            if len(covariance) != len(observation) or any(len(x) != len(observation) for x in covariance):
                raise PFARuntimeException("misaligned covariance", self.errcodeBase + 2, self.name, pos)
            x = np().array([(o - p) for o, p in zip(observation, prediction)])
            C = np().array(covariance)
        else:
            if len(observation) < 1:
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if (set(observation.keys()) != set(prediction.keys())):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            # use observation keys throughout
            keys = list(observation.keys())
            try:
                x = np().array([observation[key] - prediction[key] for key in keys])
            except:
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            C = np().empty((len(keys), len(keys)))
            try:
                for i,k1 in enumerate(keys):
                    for j,k2 in enumerate(keys):
                        C[i,j] = float(covariance[k1][k2])
            except:
                raise PFARuntimeException("misaligned covariance", self.errcodeBase + 2, self.name, pos)
        resultSquared = float(np().linalg.solve(C, x).T.dot(x))
        if resultSquared < 0:
            return float("nan")
        else:
            return math.sqrt(resultSquared)
provide(Mahalanobis())


class UpdateChi2(LibFcn):
    name = prefix + "updateChi2"
    sig = Sigs([Sig([{"pull": P.Double()}, {"state": P.WildRecord("A", {"chi2": P.Double(), "dof": P.Int()})}], P.Wildcard("A")),
                Sig([{"pull": P.Array(P.Double())}, {"state": P.WildRecord("A", {"chi2": P.Double(), "dof": P.Int()})}], P.Wildcard("A")),
                Sig([{"pull": P.Map(P.Double())}, {"state": P.WildRecord("A", {"chi2": P.Double(), "dof": P.Int()})}], P.Wildcard("A"))])
    errcodeBase = 38040
    def __call__(self, state, scope, pos, paramTypes, pull, state_):
        if isinstance(pull, float):
            return dict(state_, chi2=(state_["chi2"] + pull*pull), dof=(state_["dof"] + 1))
        elif isinstance(pull, (tuple, list)):
            return dict(state_, chi2=(state_["chi2"] + sum([y**2 for y in pull])), dof=(state_["dof"] + 1))
        else:
            return dict(state_, chi2=(state_["chi2"] + sum([y**2 for y in list(pull.values())])), dof=(state_["dof"] + 1))
provide(UpdateChi2())

class ReducedChi2(LibFcn):
    name = prefix + "reducedChi2"
    sig = Sig([{"state": P.WildRecord("A", {"chi2": P.Double(), "dof": P.Int()})}], P.Double())
    errcodeBase = 38050
    def __call__(self, state, scope, pos, paramTypes, state_):
        return div(state_["chi2"], state_["dof"])
provide(ReducedChi2())

class Chi2Prob(LibFcn):
    name = prefix + "chi2Prob"
    sig = Sig([{"state": P.WildRecord("A", {"chi2": P.Double(), "dof": P.Int()})}], P.Double())
    errcodeBase = 38060
    def __call__(self, state, scope, pos, paramTypes, state_):
        chi2 = state_["chi2"]
        dof = state_["dof"]
        if dof < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif dof == 0:
            if chi2 > 0:
                return 1.0
            else:
                return 0.0
        elif math.isnan(chi2):
            return float("nan")
        elif math.isinf(chi2):
            if chi2 > 0:
                return 1.0
            else:
                return 0.0
        else:
            return float(Chi2Distribution(dof, self.errcodeBase + 0, self.name, pos).CDF(chi2))
provide(Chi2Prob())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import itertools

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn
from titus.util import div
from titus.lib.core import powLikeJava
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "metric."

class SimpleEuclidean(LibFcn):
    name = prefix + "simpleEuclidean"
    sig = Sig([{"x": P.Array(P.Double())}, {"y": P.Array(P.Double())}], P.Double())
    errcodeBase = 28000
    def __call__(self, state, scope, pos, paramTypes, x, y):
        if len(x) != len(y):
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        return math.sqrt(sum(((xi - yi)**2 for xi, yi in zip(x, y))))
provide(SimpleEuclidean())

class AbsDiff(LibFcn):
    name = prefix + "absDiff"
    sig = Sig([{"x": P.Double()}, {"y": P.Double()}], P.Double())
    errcodeBase = 28010
    def __call__(self, state, scope, pos, paramTypes, x, y):
        return abs(x - y)
provide(AbsDiff())

class GaussianSimilarity(LibFcn):
    name = prefix + "gaussianSimilarity"
    sig = Sig([{"x": P.Double()}, {"y": P.Double()}, {"sigma": P.Double()}], P.Double())
    errcodeBase = 28020
    def __call__(self, state, scope, pos, paramTypes, x, y, sigma):
        return math.exp(div(-math.log(2) * (x - y)**2, sigma**2))
provide(GaussianSimilarity())

class MetricWithMissingValues(LibFcn):
    def __call__(self, state, scope, pos, paramTypes, similarity, x, y, missingWeight=None):
        length = len(x)
        if missingWeight is None:
            missingWeight = [1.0] * length
        if len(y) != length or len(missingWeight) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        tally = 0.0
        numer = 0.0
        denom = 0.0
        for i in range(length):
            xi = x[i]
            yi = y[i]
            if xi is not None and yi is not None:
                if isinstance(paramTypes[1]["items"], (tuple, list)):
                    xi, = list(xi.values())
                if isinstance(paramTypes[2]["items"], (tuple, list)):
                    yi, = list(yi.values())
                tally = self.increment(tally, callfcn(state, scope, similarity, [xi, yi]))
                denom += missingWeight[i]
            numer += missingWeight[i]
        if denom == 0.0:
            return self.finalize(float("nan"))
        else:
            return self.finalize(tally * numer / denom)

class Euclidean(MetricWithMissingValues):
    name = prefix + "euclidean"
    sig = Sigs([
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}], P.Double()),
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"missingWeight": P.Array(P.Double())}], P.Double())
        ])
    errcodeBase = 28030
    def increment(self, tally, x):
        return tally + x**2
    def finalize(self, x):
        return math.sqrt(x)
provide(Euclidean())

class SquaredEuclidean(MetricWithMissingValues):
    name = prefix + "squaredEuclidean"
    sig = Sigs([
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}], P.Double()),
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"missingWeight": P.Array(P.Double())}], P.Double())
        ])
    errcodeBase = 28040
    def increment(self, tally, x):
        return tally + x**2
    def finalize(self, x):
        return x
provide(SquaredEuclidean())

class Chebyshev(MetricWithMissingValues):
    name = prefix + "chebyshev"
    sig = Sigs([
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}], P.Double()),
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"missingWeight": P.Array(P.Double())}], P.Double())
        ])
    errcodeBase = 28050
    def increment(self, tally, x):
        return max(tally, x)
    def finalize(self, x):
        return x
provide(Chebyshev())

class Taxicab(MetricWithMissingValues):
    name = prefix + "taxicab"
    sig = Sigs([
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}], P.Double()),
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"missingWeight": P.Array(P.Double())}], P.Double())
        ])
    errcodeBase = 28060
    def increment(self, tally, x):
        return tally + x
    def finalize(self, x):
        return x
provide(Taxicab())

class Minkowski(LibFcn):
    name = prefix + "minkowski"
    sig = Sigs([
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"p": P.Double()}], P.Double()),
        Sig([{"similarity": P.Fcn([P.Wildcard("A"), P.Wildcard("B")], P.Double())}, {"x": P.Array(P.Union([P.Null(), P.Wildcard("A")]))}, {"y": P.Array(P.Union([P.Null(), P.Wildcard("B")]))}, {"p": P.Double()}, {"missingWeight": P.Array(P.Double())}], P.Double())
        ])
    errcodeBase = 28070
    def __call__(self, state, scope, pos, paramTypes, similarity, x, y, p, missingWeight=None):
        length = len(x)
        if missingWeight is None:
            missingWeight = [1.0] * length
        if len(y) != length or len(missingWeight) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        if math.isnan(p) or p <= 0:
            raise PFARuntimeException("Minkowski parameter p must be positive", self.errcodeBase + 1, self.name, pos)
        tally = 0.0
        numer = 0.0
        denom = 0.0
        if math.isinf(p):
            for i in range(length):
                xi = x[i]
                yi = y[i]
                if xi is not None and yi is not None:
                    if isinstance(paramTypes[1]["items"], (tuple, list)):
                        xi, = list(xi.values())
                    if isinstance(paramTypes[2]["items"], (tuple, list)):
                        yi, = list(yi.values())
                    z = callfcn(state, scope, similarity, [xi, yi])
                    if z > tally:
                        tally = z
                    denom += missingWeight[i]
                numer += missingWeight[i]
            if denom == 0.0:
                return float("nan")
            else:
                return tally * numer / denom
        else:
            for i in range(length):
                xi = x[i]
                yi = y[i]
                if xi is not None and yi is not None:
                    if isinstance(paramTypes[1]["items"], (tuple, list)):
                        xi, = list(xi.values())
                    if isinstance(paramTypes[2]["items"], (tuple, list)):
                        yi, = list(yi.values())
                    tally += powLikeJava(callfcn(state, scope, similarity, [xi, yi]), p)
                    denom += missingWeight[i]
                numer += missingWeight[i]
            if denom == 0.0:
                return float("nan")
            else:
                return powLikeJava(tally * numer / denom, 1.0/p)
provide(Minkowski())

class BinaryMetric(LibFcn):
    def countPairs(self, x, y):
        a00 = 0
        a01 = 0
        a10 = 0
        a11 = 0
        for xi, yi in zip(x, y):
            if xi and yi: a11 += 1
            if xi and not yi: a10 += 1
            if not xi and yi: a01 += 1
            if not xi and not yi: a00 += 1
        return a00, a01, a10, a11

class SimpleMatching(BinaryMetric):
    name = prefix + "simpleMatching"
    sig = Sig([{"x": P.Array(P.Boolean())}, {"y": P.Array(P.Boolean())}], P.Double())
    errcodeBase = 28080
    def __call__(self, state, scope, pos, paramTypes, x, y):
        length = len(x)
        if len(y) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        a00, a01, a10, a11 = self.countPairs(x, y)
        return div((a11 + a00), (a11 + a10 + a01 + a00))
provide(SimpleMatching())

class Jaccard(BinaryMetric):
    name = prefix + "jaccard"
    sig = Sig([{"x": P.Array(P.Boolean())}, {"y": P.Array(P.Boolean())}], P.Double())
    errcodeBase = 28090
    def __call__(self, state, scope, pos, paramTypes, x, y):
        length = len(x)
        if len(y) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        a00, a01, a10, a11 = self.countPairs(x, y)
        return div(a11, (a11 + a10 + a01))
provide(Jaccard())

class Tanimoto(BinaryMetric):
    name = prefix + "tanimoto"
    sig = Sig([{"x": P.Array(P.Boolean())}, {"y": P.Array(P.Boolean())}], P.Double())
    errcodeBase = 28100
    def __call__(self, state, scope, pos, paramTypes, x, y):
        length = len(x)
        if len(y) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        a00, a01, a10, a11 = self.countPairs(x, y)
        return div((a11 + a00), (a11 + 2*(a10 + a01) + a00))
provide(Tanimoto())

class BinarySimilarity(BinaryMetric):
    name = prefix + "binarySimilarity"
    sig = Sig([{"x": P.Array(P.Boolean())}, {"y": P.Array(P.Boolean())}, {"c00": P.Double()}, {"c01": P.Double()}, {"c10": P.Double()}, {"c11": P.Double()}, {"d00": P.Double()}, {"d01": P.Double()}, {"d10": P.Double()}, {"d11": P.Double()}], P.Double())
    errcodeBase = 28110
    def __call__(self, state, scope, pos, paramTypes, x, y, c00, c01, c10, c11, d00, d01, d10, d11):
        length = len(x)
        if len(y) != length:
            raise PFARuntimeException("dimensions of vectors do not match", self.errcodeBase + 0, self.name, pos)
        a00, a01, a10, a11 = self.countPairs(x, y)
        return div((c11*a11 + c10*a10 + c01*a01 + c00*a00), (d11*a11 + d10*a10 + d01*a01 + d00*a00))
provide(BinarySimilarity())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import div, callfcn
from titus.lib.core import INT_MIN_VALUE
from titus.lib.core import INT_MAX_VALUE
import titus.P as P
from functools import reduce

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "stat.sample."

#################################################################### 

class UpdateMeanVariance(LibFcn):
    def genpy(self, paramTypes, args, pos):
        hasMean = False
        hasVariance = False
        for x in self._getRecord(paramTypes[2]).fields:
            if x.name == "mean":
                if not x.avroType.accepts(AvroDouble()):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"mean\" field is not a double: " + str(x.avroType), None)
                hasMean = True
            elif x.name == "variance":
                if not x.avroType.accepts(AvroDouble()):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"variance\" field is not a double: " + str(x.avroType), None)
                hasVariance = True
        if (hasMean, hasVariance) == (False, False):
            level = 0
        elif (hasMean, hasVariance) == (True, False):
            level = 1
        elif (hasMean, hasVariance) == (True, True):
            level = 2
        elif (hasMean, hasVariance) == (False, True):
            raise PFASemanticException(prefix + "update with \"variance\" must also have \"mean\" in the state record type", None)
        return "self.f[{0}]({1}, {2})".format(repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args), level)

class Update(UpdateMeanVariance):
    name = prefix + "update"
    sig = Sig([{"x": P.Double()}, {"w": P.Double()}, {"state": P.WildRecord("A", {"count": P.Double()})}], P.Wildcard("A"))
    errcodeBase = 14000
    def _getRecord(self, paramType):
        if isinstance(paramType, AvroUnion):
            for t in paramType.types:
                if not isinstance(t, AvroNull):
                    return t
        else:
            return paramType

    def __call__(self, state, scope, pos, paramTypes, x, w, theState, level):
        originalCount = theState["count"]
        count = originalCount + w
        if level == 0:
            return dict(theState, count=count)
        else:
            mean = theState["mean"]
            delta = x - mean
            shift = div(delta * w, count)
            mean += shift
            if level == 1:
                return dict(theState, count=count, mean=mean)
            else:
                varianceTimesCount = theState["variance"] * originalCount
                varianceTimesCount += originalCount * delta * shift
                return dict(theState, count=count, mean=mean, variance=div(varianceTimesCount, count))

provide(Update())

class UpdateCovariance(LibFcn):
    name = prefix + "updateCovariance"
    sig = Sigs([Sig([{"x": P.Array(P.Double())}, {"w": P.Double()}, {"state": P.WildRecord("A", {"count": P.Double(), "mean": P.Array(P.Double()), "covariance": P.Array(P.Array(P.Double()))})}], P.Wildcard("A")),
                Sig([{"x": P.Map(P.Double())}, {"w": P.Double()}, {"state": P.WildRecord("A", {"count": P.Map(P.Map(P.Double())), "mean": P.Map(P.Double()), "covariance": P.Map(P.Map(P.Double()))})}], P.Wildcard("A"))])

    errcodeBase = 14010
    def __call__(self, state, scope, pos, paramTypes, x, w, theState):
        size = len(x)
        if size < 2:
            raise PFARuntimeException("too few components", self.errcodeBase + 1, self.name, pos)

        if paramTypes[0]["type"] == "map":
            oldCount = theState["count"]
            oldMean = theState["mean"]
            oldCovariance = theState["covariance"]

            countKeys = set(oldCount.keys()).union(reduce(lambda a, b: a.union(b), (set(v.keys()) for v in list(oldCount.values())), set()))
            covarKeys = set(oldCovariance.keys()).union(reduce(lambda a, b: a.union(b), (set(v.keys()) for v in list(oldCovariance.values())), set()))
            keys = set(x.keys()).union(countKeys).union(covarKeys)

            newCount = {}
            for i in keys:
                row = {}
                for j in keys:
                    old = oldCount.get(i, {}).get(j, 0.0)
                    if (i in x) and (j in x):
                        row[j] = old + w
                    else:
                        row[j] = old
                newCount[i] = row

            newMean = {}
            for i in keys:
                old = oldMean.get(i, 0.0)
                if i in x:
                    newMean[i] = old + div((x[i] - old) * w, newCount[i][i])
                else:
                    newMean[i] = old

            newCovariance = {}
            for i in keys:
                row = {}
                for j in keys:
                    oldCov = oldCovariance.get(i, {}).get(j, 0.0)
                    if (i in x) and (j in x):
                        oldC = oldCount.get(i, {}).get(j, 0.0)
                        oldMi = oldMean.get(i, 0.0)
                        oldMj = oldMean.get(j, 0.0)
                        row[j] = div(((oldCov*oldC) + div((x[i] - oldMi) * (x[j] - oldMj) * w*oldC, newCount[i][j])), newCount[i][j])
                    else:
                        row[j] = oldCov
                newCovariance[i] = row

            return dict(theState, count=newCount, mean=newMean, covariance=newCovariance)

        else:
            oldCount = theState["count"]
            oldMean = theState["mean"]
            oldCovariance = theState["covariance"]

            if (size != len(oldMean) or size != len(oldCovariance) or any(size != len(xi) for xi in oldCovariance)):
                raise PFARuntimeException("unequal length arrays", self.errcodeBase + 2, self.name, pos)

            newCount = oldCount + w
            newMean = [oldm + div((x[i] - oldm) * w, newCount) for i, oldm in enumerate(oldMean)]
            newCovariance = []
            for i in range(size):
                row = []
                for j in range(size):
                    row.append(div((oldCovariance[i][j]*oldCount) + div((x[i] - oldMean[i]) * (x[j] - oldMean[j]) * w*oldCount, newCount), newCount))
                newCovariance.append(row)

            return dict(theState, count=newCount, mean=newMean, covariance=newCovariance)

provide(UpdateCovariance())

class UpdateWindow(UpdateMeanVariance):
    name = prefix + "updateWindow"
    sig = Sig([{"x": P.Double()}, {"w": P.Double()}, {"state": P.Array(P.WildRecord("A", {"x": P.Double(), "w": P.Double(), "count": P.Double()}))}, {"windowSize": P.Int()}], P.Array(P.Wildcard("A")))
    errcodeBase = 14020
    def _getRecord(self, paramType):
        return paramType.items

    def __call__(self, state, scope, pos, paramTypes, x, w, theState, windowSize, level):
        if windowSize < 2:
            raise PFARuntimeException("windowSize must be at least 2", self.errcodeBase + 0, self.name, pos)

        if len(theState) == 0:
            out = {}
            paramType = state.parser.getAvroType(paramTypes[2])
            for field in self._getRecord(paramType).fields:
                if field.name == "x":
                    out["x"] = x
                elif field.name == "w":
                    out["w"] = w
                elif field.name == "count":
                    out["count"] = w
                elif field.name == "mean":
                    out["mean"] = x
                elif field.name == "variance":
                    out["variance"] = 0.0
                else:
                    raise PFARuntimeException("cannot initialize unrecognized fields", self.errcodeBase + 1, self.name, pos)
            return [out]
        
        elif len(theState) >= windowSize:
            record = theState[-1]

            splitAt = len(theState) - windowSize + 1
            remove, keep = theState[:splitAt], theState[splitAt:]
            oldx = [xi["x"] for xi in remove]
            oldw = [-xi["w"] for xi in remove]

            originalCount = record["count"]
            count = originalCount + w

            count2 = count + sum(oldw)

            if level == 0:
                return keep + [dict(record, x=x, w=w, count=count2)]
            else:
                mean = record["mean"]
                delta = x - mean
                shift = div(delta * w, count)

                mean += shift

                accumulatedCount = count
                varianceCorrection = 0.0
                for ox, ow in zip(oldx, oldw):
                    accumulatedCount += ow
                    delta2 = ox - mean
                    shift2 = div(delta2 * ow, accumulatedCount)

                    mean += shift2
                    varianceCorrection += (accumulatedCount - ow) * delta2 * shift2

                if level == 1:
                    return keep + [dict(record, x=x, w=w, count=count2, mean=mean)]
                else:
                    varianceTimesCount = record["variance"] * originalCount
                    varianceTimesCount += originalCount * delta * shift

                    varianceTimesCount += varianceCorrection

                    return keep + [dict(record, x=x, w=w, count=count2, mean=mean, variance=div(varianceTimesCount, count2))]

        else:
            record = theState[-1]

            originalCount = record["count"]
            count = originalCount + w

            if level == 0:
                return theState + [dict(record, x=x, w=w, count=count)]
            else:
                mean = record["mean"]
                delta = x - mean
                shift = div(delta * w, count)
                mean += shift

                if level == 1:
                    return theState + [dict(record, x=x, w=w, count=count, mean=mean)]
                else:
                    varianceTimesCount = record["variance"] * originalCount
                    varianceTimesCount += originalCount * delta * shift

                    return theState + [dict(record, x=x, w=w, count=count, mean=mean, variance=div(varianceTimesCount, count))]

provide(UpdateWindow())

class UpdateEWMA(LibFcn):
    name = prefix + "updateEWMA"
    sig = Sig([{"x": P.Double()}, {"alpha": P.Double()}, {"state": P.WildRecord("A", {"mean": P.Double()})}], P.Wildcard("A"))
    errcodeBase = 14030
    def _getRecord(self, paramType):
        if isinstance(paramType, AvroUnion):
            for t in paramType.types:
                if not isinstance(t, AvroNull):
                    return t
        else:
            return paramType

    def genpy(self, paramTypes, args, pos):
        hasVariance = False
        for x in self._getRecord(paramTypes[2]).fields:
            if x.name == "variance":
                if not x.avroType.accepts(AvroDouble()):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"variance\" field is not a double: " + str(x.avroType), None)
                hasVariance = True
        return "self.f[{0}]({1}, {2})".format(repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args), hasVariance)

    def __call__(self, state, scope, pos, paramTypes, x, alpha, theState, hasVariance):
        if alpha < 0.0 or alpha > 1.0:
            raise PFARuntimeException("alpha out of range", self.errcodeBase + 0, self.name, pos)

        mean = theState["mean"]
        diff = x - mean
        incr = alpha * diff

        if hasVariance:
            variance = theState["variance"]
            return dict(theState, mean=(mean + incr), variance=((1.0 - alpha) * (variance + diff * incr)))
        else:
            return dict(theState, mean=(mean + incr))

provide(UpdateEWMA())

class UpdateHoltWinters(LibFcn):
    name = prefix + "updateHoltWinters"
    sig = Sig([{"x": P.Double()}, {"alpha": P.Double()}, {"beta": P.Double()}, {"state": P.WildRecord("A", {"level": P.Double(), "trend": P.Double()})}], P.Wildcard("A"))
    errcodeBase = 14040
    def __call__(self, state, scope, pos, paramTypes, x, alpha, beta, theState):
        if alpha < 0.0 or alpha > 1.0:
            raise PFARuntimeException("alpha out of range", self.errcodeBase + 0, self.name, pos)
        if beta < 0.0 or beta > 1.0:
            raise PFARuntimeException("beta out of range", self.errcodeBase + 1, self.name, pos)
        level_prev = theState["level"]
        trend_prev = theState["trend"]
        level = alpha * x + (1.0 - alpha) * (level_prev + trend_prev)
        trend = beta * (level - level_prev) + (1.0 - beta) * trend_prev
        return dict(theState, level=level, trend=trend)
provide(UpdateHoltWinters())

class UpdateHoltWintersPeriodic(LibFcn):
    name = prefix + "updateHoltWintersPeriodic"
    sig = Sig([{"x": P.Double()}, {"alpha": P.Double()}, {"beta": P.Double()}, {"gamma": P.Double()}, {"state": P.WildRecord("A", {"level": P.Double(), "trend": P.Double(), "cycle": P.Array(P.Double()), "multiplicative": P.Boolean()})}], P.Wildcard("A"))
    errcodeBase = 14050
    def __call__(self, state, scope, pos, paramTypes, x, alpha, beta, gamma, theState):
        if alpha < 0.0 or alpha > 1.0:
            raise PFARuntimeException("alpha out of range", self.errcodeBase + 0, self.name, pos)
        if beta < 0.0 or beta > 1.0:
            raise PFARuntimeException("beta out of range", self.errcodeBase + 1, self.name, pos)
        if gamma < 0.0 or gamma > 1.0:
            raise PFARuntimeException("gamma out of range", self.errcodeBase + 2, self.name, pos)

        level_prev = theState["level"]
        trend_prev = theState["trend"]
        cycle_unrotated = theState["cycle"]
        if len(cycle_unrotated) == 0:
            raise PFARuntimeException("empty cycle", self.errcodeBase + 3, self.name, pos)
        cycle_rotated = cycle_unrotated[1:] + [cycle_unrotated[0]]
        cycle_prev = cycle_rotated[0]

        if theState["multiplicative"]:
            level = div(alpha * x, cycle_prev) + (1.0 - alpha) * (level_prev + trend_prev)
            trend = beta * (level - level_prev) + (1.0 - beta) * trend_prev
            cycle = div(gamma * x, level) + (1.0 - gamma) * cycle_prev
        else:
            level = alpha * (x - cycle_prev) + (1.0 - alpha) * (level_prev + trend_prev)
            trend = beta * (level - level_prev) + (1.0 - beta) * trend_prev
            cycle = gamma * (x - level) + (1.0 - gamma) * cycle_prev

        return dict(theState, level=level, trend=trend, cycle=([cycle] + cycle_rotated[1:]))

provide(UpdateHoltWintersPeriodic())

class Forecast1HoltWinters(LibFcn):
    name = prefix + "forecast1HoltWinters"
    sig = Sig([{"state": P.WildRecord("A", {"level": P.Double(), "trend": P.Double()})}], P.Double())

    errcodeBase = 14060
    def genpy(self, paramTypes, args, pos):
        hasCycle = False
        hasMultiplicative = False
        for x in paramTypes[0].fields:
            if x.name == "cycle":
                if not x.avroType.accepts(AvroArray(AvroDouble())):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"cycle\" field is not an array of double: " + str(x.avroType), None)
                hasCycle = True
            elif x.name == "multiplicative":
                if not x.avroType.accepts(AvroBoolean()):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"multiplicative\" field is not a boolean: " + str(x.avroType), None)
                hasMultiplicative = True

        if hasCycle ^ hasMultiplicative:
            raise PFASemanticException(self.name + " is being given a state record type with a \"cycle\" but no \"multiplicative\" or vice-versa", None)

        return "self.f[{0}]({1}, {2})".format(repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args), hasCycle)

    def __call__(self, state, scope, pos, paramTypes, theState, hasPeriodic):
        level = theState["level"]
        trend = theState["trend"]

        if not hasPeriodic:
            return level + trend
        else:
            cycle = theState["cycle"]
            L = len(cycle)
            if L == 0:
                raise PFARuntimeException("empty cycle", self.errcodeBase + 0, self.name, pos)

            if theState["multiplicative"]:
                return (level + trend) * cycle[1 % L]
            else:
                return level + trend + cycle[1 % L]

provide(Forecast1HoltWinters())

class ForecastHoltWinters(LibFcn):
    name = prefix + "forecastHoltWinters"
    sig = Sig([{"n": P.Int()}, {"state": P.WildRecord("A", {"level": P.Double(), "trend": P.Double()})}], P.Array(P.Double()))

    errcodeBase = 14070
    def genpy(self, paramTypes, args, pos):
        hasCycle = False
        hasMultiplicative = False
        for x in paramTypes[1].fields:
            if x.name == "cycle":
                if not x.avroType.accepts(AvroArray(AvroDouble())):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"cycle\" field is not an array of double: " + str(x.avroType), None)
                hasCycle = True
            elif x.name == "multiplicative":
                if not x.avroType.accepts(AvroBoolean()):
                    raise PFASemanticException(self.name + " is being given a state record type in which the \"multiplicative\" field is not a boolean: " + str(x.avroType), None)
                hasMultiplicative = True

        if hasCycle ^ hasMultiplicative:
            raise PFASemanticException(self.name + " is being given a state record type with a \"cycle\" but no \"multiplicative\" or vice-versa", None)

        return "self.f[{0}]({1}, {2})".format(repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args), hasCycle)

    def __call__(self, state, scope, pos, paramTypes, n, theState, hasPeriodic):
        level = theState["level"]
        trend = theState["trend"]

        if not hasPeriodic:
            return [level + i*trend for i in range(1, n + 1)]
        else:
            cycle = theState["cycle"]
            L = len(cycle)
            if L == 0:
                raise PFARuntimeException("empty cycle", self.errcodeBase + 0, self.name, pos)

            if theState["multiplicative"]:
                return [(level + i*trend) * cycle[i % L] for i in range(1, n + 1)]
            else:
                return [level + i*trend + cycle[i % L] for i in range(1, n + 1)]

provide(ForecastHoltWinters())

class FillHistogram(LibFcn):
    name = prefix + "fillHistogram"
    sig = Sigs([Sig([{"x": P.Double()}, {"w": P.Double()}, {"histogram": P.WildRecord("A", {"numbins": P.Int(), "low": P.Double(), "high": P.Double(), "values": P.Array(P.Double())})}], P.Wildcard("A")),
                Sig([{"x": P.Double()}, {"w": P.Double()}, {"histogram": P.WildRecord("A", {"low": P.Double(), "binsize": P.Double(), "values": P.Array(P.Double())})}], P.Wildcard("A")),
                Sig([{"x": P.Double()}, {"w": P.Double()}, {"histogram": P.WildRecord("A", {"ranges": P.Array(P.Array(P.Double())), "values": P.Array(P.Double())})}], P.Wildcard("A"))])

    errcodeBase = 14080
    def genpy(self, paramTypes, args, pos):
        def has(name, avroType):
            for x in paramTypes[2].fields:
                if x.name == name:
                    if not x.avroType.accepts(avroType):
                        raise PFASemanticException("{0} is being given a record type in which the \"{1}\" field is not {2}: {3}".format(self.name, name, avroType, x.avroType), None)
                    return True
            return False

        method0 = has("numbins", AvroInt()) and has("low", AvroDouble()) and has("high", AvroDouble())
        method1 = has("low", AvroDouble()) and has("binsize", AvroDouble())
        method2 = has("ranges", AvroArray(AvroArray(AvroDouble())))

        if       method0 and not method1 and not method2:
            method = 0
        elif not method0 and     method1 and not method2:
            method = 1
        elif not method0 and not method1 and     method2:
            method = 2
        else:
            raise PFASemanticException(prefix + "fillHistogram must have \"numbins\", \"low\", \"high\" xor it must have \"low\", \"binsize\" xor it must have \"ranges\", but not any other combination of these fields.", None)

        hasUnderflow = has("underflow", AvroDouble())
        hasOverflow = has("overflow", AvroDouble())
        hasNanflow = has("nanflow", AvroDouble())
        hasInfflow = has("infflow", AvroDouble())

        return "self.f[{0}]({1}, {2}, {3}, {4}, {5}, {6})".format(repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args), method, hasUnderflow, hasOverflow, hasNanflow, hasInfflow)

    def updateHistogram(self, w, histogram, newValues, hasUnderflow, hasOverflow, hasNanflow, hasInfflow, underflow, overflow, nanflow, infflow):
        updator = {"values": newValues}
        if (hasUnderflow):
            updator["underflow"] = histogram["underflow"] + (w if underflow else 0.0)
        if (hasOverflow):
            updator["overflow"] = histogram["overflow"] + (w if overflow else 0.0)
        if (hasNanflow):
            updator["nanflow"] = histogram["nanflow"] + (w if nanflow else 0.0)
        if (hasInfflow):
            updator["infflow"] = histogram["infflow"] + (w if infflow else 0.0)
        return dict(histogram, **updator)

    def __call__(self, state, scope, pos, paramTypes, x, w, histogram, method, hasUnderflow, hasOverflow, hasNanflow, hasInfflow):
        values = histogram["values"]

        if method == 0:
            numbins, low, high = histogram["numbins"], histogram["low"], histogram["high"]

            if numbins < 1:
                raise PFARuntimeException("bad histogram scale", self.errcodeBase + 2, self.name, pos)
            if len(values) != numbins:
                raise PFARuntimeException("wrong histogram size", self.errcodeBase + 0, self.name, pos)
            if low >= high or math.isnan(low) or math.isinf(low) or math.isnan(high) or math.isinf(high):
                raise PFARuntimeException("bad histogram range", self.errcodeBase + 1, self.name, pos)

            if hasInfflow and math.isinf(x):
                underflow, overflow, nanflow, infflow = False, False, False, True
            elif math.isnan(x):
                underflow, overflow, nanflow, infflow = False, False, True, False
            elif x >= high:
                underflow, overflow, nanflow, infflow = False, True, False, False
            elif x < low:
                underflow, overflow, nanflow, infflow = True, False, False, False
            else:
                underflow, overflow, nanflow, infflow = False, False, False, False

            if not underflow and not overflow and not nanflow and not infflow:
                try:
                    indexFloat = math.floor((x - low) / (high - low) * numbins)
                except ZeroDivisionError:
                    index = 0
                else:
                    if math.isinf(indexFloat):
                        if indexFloat > 0:
                            index = INT_MAX_VALUE
                        else:
                            index = INT_MIN_VALUE
                    else:
                        index = int(indexFloat)
                newValues = list(values)
                newValues[index] = newValues[index] + w
            else:
                newValues = values

            return self.updateHistogram(w, histogram, newValues, hasUnderflow, hasOverflow, hasNanflow, hasInfflow, underflow, overflow, nanflow, infflow)

        elif method == 1:
            low, binsize = histogram["low"], histogram["binsize"]

            if binsize <= 0.0 or math.isnan(binsize) or math.isinf(binsize):
                raise PFARuntimeException("bad histogram scale", self.errcodeBase + 2, self.name, pos)

            if hasInfflow and math.isinf(x):
                underflow, overflow, nanflow, infflow = False, False, False, True
            elif math.isnan(x):
                underflow, overflow, nanflow, infflow = False, False, True, False
            elif math.isinf(x) and x > 0.0:
                underflow, overflow, nanflow, infflow = False, True, False, False
            elif x < low:
                underflow, overflow, nanflow, infflow = True, False, False, False
            else:
                underflow, overflow, nanflow, infflow = False, False, False, False
                
            if not underflow and not overflow and not nanflow and not infflow:
                currentHigh = low + binsize * len(values)
                try:
                    index = int(math.floor((x - low) / (currentHigh - low) * len(values)))
                except ZeroDivisionError:
                    index = 0
                if index < len(values):
                    newValues = list(values)
                    newValues[index] = newValues[index] + w
                else:
                    newValues = values + [0.0] * (index - len(values)) + [w]

            else:
                newValues = values

            return self.updateHistogram(w, histogram, newValues, hasUnderflow, hasOverflow, hasNanflow, hasInfflow, underflow, overflow, nanflow, infflow)

        elif method == 2:
            ranges = histogram["ranges"]

            if len(values) != len(ranges):
                raise PFARuntimeException("wrong histogram size", self.errcodeBase + 0, self.name, pos)

            if any(len(x) != 2 or x[0] >= x[1] or math.isnan(x[0]) or math.isinf(x[0]) or math.isnan(x[1]) or math.isinf(x[1]) for x in ranges):
                raise PFARuntimeException("bad histogram ranges", self.errcodeBase + 3, self.name, pos)

            isInfinite = math.isinf(x)
            isNan = math.isnan(x)

            newValues = list(values)
            hitOne = False

            if not isInfinite and not isNan:
                for index, rang in enumerate(ranges):
                    low = rang[0]
                    high = rang[1]

                    if low == high and x == low:
                        newValues[index] = newValues[index] + w
                        hitOne = True

                    elif x >= low and x < high:
                        newValues[index] = newValues[index] + w
                        hitOne = True

            if hasInfflow and isInfinite:
                underflow, overflow, nanflow, infflow = False, False, False, True
            elif isNan:
                underflow, overflow, nanflow, infflow = False, False, True, False
            elif not hitOne:
                underflow, overflow, nanflow, infflow = True, False, False, False
            else:
                underflow, overflow, nanflow, infflow = False, False, False, False

            return self.updateHistogram(w, histogram, newValues, hasUnderflow, hasOverflow, hasNanflow, hasInfflow, underflow, overflow, nanflow, infflow)

provide(FillHistogram())

class FillHistogram2d(LibFcn):
    name = prefix + "fillHistogram2d"
    sig = Sig([{"x": P.Double()}, {"y": P.Double()}, {"w": P.Double()}, {"histogram": P.WildRecord("A", {"xnumbins": P.Int(), "xlow": P.Double(), "xhigh": P.Double(), "ynumbins": P.Int(), "ylow": P.Double(), "yhigh": P.Double(), "values": P.Array(P.Array(P.Double()))})}], P.Wildcard("A"))

    errcodeBase = 14090
    def genpy(self, paramTypes, args, pos):
        def has(name):
            for x in paramTypes[3].fields:
                if x.name == name:
                    if not x.avroType.accepts(AvroDouble()):
                        raise PFASemanticException("{0} is being given a record type in which the \"{1}\" field is not a double: {2}".format(self.name, name, x.avroType), None)
                    return True
            return False

        return "self.f[{0}]({1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9}, {10}, {11})".format(
            repr(self.name), ", ".join(["state", "scope", repr(pos), repr(paramTypes)] + args),
            has("underunderflow"),
            has("undermidflow"),
            has("underoverflow"),
            has("midunderflow"),
            has("midoverflow"),
            has("overunderflow"),
            has("overmidflow"),
            has("overoverflow"),
            has("nanflow"),
            has("infflow"))

    def __call__(self, state, scope, pos, paramTypes, x, y, w, histogram,
                 hasUnderunderflow, hasUndermidflow, hasUnderoverflow,
                 hasMidunderflow,                    hasMidoverflow,
                 hasOverunderflow,  hasOvermidflow,  hasOveroverflow,
                 hasNanflow, hasInfflow):

        values = histogram["values"]
        xnumbins = histogram["xnumbins"]
        xlow = histogram["xlow"]
        xhigh = histogram["xhigh"]
        ynumbins = histogram["ynumbins"]
        ylow = histogram["ylow"]
        yhigh = histogram["yhigh"]

        if xnumbins < 1 or ynumbins < 1:
            raise PFARuntimeException("bad histogram scale", self.errcodeBase + 2, self.name, pos)
        if len(values) != xnumbins or any(len(x) != ynumbins for x in values):
            raise PFARuntimeException("wrong histogram size", self.errcodeBase + 0, self.name, pos)
        if xlow >= xhigh or ylow >= yhigh or math.isnan(xlow) or math.isinf(xlow) or math.isnan(xhigh) or math.isinf(xhigh) or math.isnan(ylow) or math.isinf(ylow) or math.isnan(yhigh) or math.isinf(yhigh):
            raise PFARuntimeException("bad histogram range", self.errcodeBase + 1, self.name, pos)

        if math.isnan(x) or math.isnan(y):  # do nan check first: nan wins over inf
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, False, False, False, True, False
        elif hasInfflow and (math.isinf(x) or math.isinf(y)):
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, False, False, False, False, True
        elif x >= xhigh and y >= yhigh:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, False, False, True, False, False
        elif x >= xhigh and y >= ylow and y < yhigh:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, False, True, False, False, False
        elif x >= xhigh and y < ylow:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, True, False, False, False, False
        elif x >= xlow and x < xhigh and y >= yhigh:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, True, False, False, False, False, False
        elif x >= xlow and x < xhigh and y < ylow:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, True, False, False, False, False, False, False
        elif x < xlow and y >= yhigh:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, True, False, False, False, False, False, False, False
        elif x < xlow and y >= ylow and y < yhigh:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, True, False, False, False, False, False, False, False, False
        elif x < xlow and y < ylow:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = True, False, False, False, False, False, False, False, False, False
        else:
            underunderflow, undermidflow, underoverflow, midunderflow, midoverflow, overunderflow, overmidflow, overoverflow, nanflow, infflow = False, False, False, False, False, False, False, False, False, False

        if not underunderflow and not undermidflow and not underoverflow and not midunderflow and not midoverflow and not overunderflow and not overmidflow and not overoverflow and not nanflow and not infflow:
            try:
                xindex = int(math.floor((x - xlow) / (xhigh - xlow) * xnumbins))
            except ZeroDivisionError:
                xindex = 0
            try:
                yindex = int(math.floor((y - ylow) / (yhigh - ylow) * ynumbins))
            except ZeroDivisionError:
                yindex = 0
            newValues = [list(x) for x in values]
            newValues[xindex][yindex] = newValues[xindex][yindex] + w
        else:
            newValues = values

        updator = {"values": newValues}
        if hasUnderunderflow:
            updator["underunderflow"] = histogram["underunderflow"] + (w if underunderflow else 0.0)
        if hasUndermidflow:
            updator["undermidflow"] = histogram["undermidflow"] + (w if undermidflow else 0.0)
        if hasUnderoverflow:
            updator["underoverflow"] = histogram["underoverflow"] + (w if underoverflow else 0.0)
        if hasMidunderflow:
            updator["midunderflow"] = histogram["midunderflow"] + (w if midunderflow else 0.0)
        if hasMidoverflow:
            updator["midoverflow"] = histogram["midoverflow"] + (w if midoverflow else 0.0)
        if hasOverunderflow:
            updator["overunderflow"] = histogram["overunderflow"] + (w if overunderflow else 0.0)
        if hasOvermidflow:
            updator["overmidflow"] = histogram["overmidflow"] + (w if overmidflow else 0.0)
        if hasOveroverflow:
            updator["overoverflow"] = histogram["overoverflow"] + (w if overoverflow else 0.0)
        if hasNanflow:
            updator["nanflow"] = histogram["nanflow"] + (w if nanflow else 0.0)
        if hasInfflow:
            updator["infflow"] = histogram["infflow"] + (w if infflow else 0.0)

        return dict(histogram, **updator)

provide(FillHistogram2d())

class FillCounter(LibFcn):
    name = prefix + "fillCounter"
    sig = Sig([{"x": P.String()}, {"w": P.Double()}, {"counter": P.WildRecord("A", {"values": P.Map(P.Double())})}], P.Wildcard("A"))
    errcodeBase = 14100
    def __call__(self, state, scope, pos, paramTypes, x, w, counter):
        oldmap = counter["values"]
        newmap = dict(oldmap, **{x: oldmap.get(x, 0.0) + w})
        return dict(counter, values=newmap)

provide(FillCounter())

class TopN(LibFcn):
    name = prefix + "topN"
    sig = Sig([{"x": P.Wildcard("A")}, {"top": P.Array(P.Wildcard("A"))}, {"n": P.Int()}, {"lessThan": P.Fcn([P.Wildcard("A"), P.Wildcard("A")], P.Boolean())}], P.Array(P.Wildcard("A")))
    errcodeBase = 14110
    def __call__(self, state, scope, pos, paramTypes, x, top, n, lessThan):
        if n <= 0:
            return []
        else:
            index = 0
            for best in top:
                if callfcn(state, scope, lessThan, [best, x]):
                    break
                index += 1
            if index == len(top):
                out = top + [x]
            else:
                above, below = top[:index], top[index:]
                out = above + [x] + below
            return out[:n]

provide(TopN())
